# distutils: include_dirs = EMBREE_INC_DIR
# distutils: library_dirs = EMBREE_LIB_DIR
# distutils: libraries = EMBREE_LIBS
# distutils: language = c++
"""
This file contains the ElementMesh classes, which represent the target that the
rays will be cast at when rendering finite element data. This class handles
the interface between the internal representation of the mesh and the pyembree
representation.

Note - this file is only used for the Embree-accelerated ray-tracer.

"""


import numpy as np

cimport cython
cimport numpy as np
cimport pyembree.rtcore as rtc
cimport pyembree.rtcore_geometry as rtcg
cimport pyembree.rtcore_geometry_user as rtcgu
cimport pyembree.rtcore_ray as rtcr
from libc.math cimport fmax, sqrt
from libc.stdlib cimport free, malloc
from mesh_intersection cimport (
    patchBoundsFunc,
    patchIntersectFunc,
    tet_patchBoundsFunc,
    tet_patchIntersectFunc,
)
from mesh_samplers cimport sample_hex, sample_tetra, sample_wedge
from mesh_traversal cimport YTEmbreeScene
from pyembree.rtcore cimport Triangle, Vec3f, Vertex

from yt.utilities.exceptions import YTElementTypeNotRecognized


cdef extern from "mesh_triangulation.h":
    enum:
        MAX_NUM_TRI

    int HEX_NV
    int HEX_NT
    int TETRA_NV
    int TETRA_NT
    int WEDGE_NV
    int WEDGE_NT
    int triangulate_hex[MAX_NUM_TRI][3]
    int triangulate_tetra[MAX_NUM_TRI][3]
    int triangulate_wedge[MAX_NUM_TRI][3]
    int hex20_faces[6][8]
    int tet10_faces[4][6]


cdef class LinearElementMesh:
    r'''

    This creates a 1st-order mesh to be ray-traced with embree.
    Currently, we handle non-triangular mesh types by converting them
    to triangular meshes. This class performs this transformation.
    Currently, this is implemented for hexahedral and tetrahedral
    meshes.

    Parameters
    ----------

    scene : EmbreeScene
        This is the scene to which the constructed polygons will be
        added.
    vertices : a np.ndarray of floats.
        This specifies the x, y, and z coordinates of the vertices in
        the polygon mesh. This should either have the shape
        (num_vertices, 3). For example, vertices[2][1] should give the
        y-coordinate of the 3rd vertex in the mesh.
    indices : a np.ndarray of ints
        This should either have the shape (num_elements, 4) or
        (num_elements, 8) for tetrahedral and hexahedral meshes,
        respectively. For tetrahedral meshes, each element will
        be represented by four triangles in the scene. For hex meshes,
        each element will be represented by 12 triangles, 2 for each
        face. For hex meshes, we assume that the node ordering is as
        defined here:
        http://homepages.cae.wisc.edu/~tautges/papers/cnmev3.pdf

    '''

    cdef Vertex* vertices
    cdef Triangle* indices
    cdef unsigned int mesh
    cdef double* field_data
    cdef rtcg.RTCFilterFunc filter_func
    # triangles per element, vertices per element, and field points per
    # element, respectively:
    cdef int tpe, vpe, fpe
    cdef int[MAX_NUM_TRI][3] tri_array
    cdef int* element_indices
    cdef MeshDataContainer datac

    def __init__(self, YTEmbreeScene scene,
                 np.ndarray vertices,
                 np.ndarray indices,
                 np.ndarray data):

        # We need to figure out what kind of elements we've been handed.
        if indices.shape[1] == 8:
            self.vpe = HEX_NV
            self.tpe = HEX_NT
            self.tri_array = triangulate_hex
        elif indices.shape[1] == 6:
            self.vpe = WEDGE_NV
            self.tpe = WEDGE_NT
            self.tri_array = triangulate_wedge
        elif indices.shape[1] == 4:
            self.vpe = TETRA_NV
            self.tpe = TETRA_NT
            self.tri_array = triangulate_tetra
        else:
            raise YTElementTypeNotRecognized(vertices.shape[1],
                                             indices.shape[1])

        self._build_from_indices(scene, vertices, indices)
        self._set_field_data(scene, data)
        self._set_sampler_type(scene)

    cdef void _build_from_indices(self, YTEmbreeScene scene,
                                  np.ndarray vertices_in,
                                  np.ndarray indices_in):
        cdef int i, j
        cdef int nv = vertices_in.shape[0]
        cdef int ne = indices_in.shape[0]
        cdef int nt = self.tpe*ne

        cdef unsigned int mesh = rtcg.rtcNewTriangleMesh(scene.scene_i,
                    rtcg.RTC_GEOMETRY_STATIC, nt, nv, 1)

        # first just copy over the vertices
        cdef Vertex* vertices = <Vertex*> malloc(nv * sizeof(Vertex))
        for i in range(nv):
            vertices[i].x = vertices_in[i, 0]
            vertices[i].y = vertices_in[i, 1]
            vertices[i].z = vertices_in[i, 2]
        rtcg.rtcSetBuffer(scene.scene_i, mesh, rtcg.RTC_VERTEX_BUFFER,
                          vertices, 0, sizeof(Vertex))

        # now build up the triangles
        cdef Triangle* triangles = <Triangle*> malloc(nt * sizeof(Triangle))
        for i in range(ne):
            for j in range(self.tpe):
                triangles[self.tpe*i+j].v0 = indices_in[i][self.tri_array[j][0]]
                triangles[self.tpe*i+j].v1 = indices_in[i][self.tri_array[j][1]]
                triangles[self.tpe*i+j].v2 = indices_in[i][self.tri_array[j][2]]
        rtcg.rtcSetBuffer(scene.scene_i, mesh, rtcg.RTC_INDEX_BUFFER,
                          triangles, 0, sizeof(Triangle))

        cdef int* element_indices = <int *> malloc(ne * self.vpe * sizeof(int))
        for i in range(ne):
            for j in range(self.vpe):
                element_indices[i*self.vpe + j] = indices_in[i][j]

        self.element_indices = element_indices
        self.vertices = vertices
        self.indices = triangles
        self.mesh = mesh

    cdef void _set_field_data(self, YTEmbreeScene scene,
                              np.ndarray data_in):

        cdef int ne = data_in.shape[0]
        self.fpe = data_in.shape[1]

        cdef double* field_data = <double *> malloc(ne * self.fpe * sizeof(double))

        for i in range(ne):
            for j in range(self.fpe):
                field_data[i*self.fpe+j] = data_in[i][j]

        self.field_data = field_data

        cdef MeshDataContainer datac
        datac.vertices = self.vertices
        datac.indices = self.indices
        datac.field_data = self.field_data
        datac.element_indices = self.element_indices
        datac.tpe = self.tpe
        datac.vpe = self.vpe
        datac.fpe = self.fpe
        self.datac = datac

        rtcg.rtcSetUserData(scene.scene_i, self.mesh, &self.datac)

    cdef void _set_sampler_type(self, YTEmbreeScene scene):

        if self.vpe == 4:
            self.filter_func = <rtcg.RTCFilterFunc> sample_tetra
        elif self.vpe == 6:
            self.filter_func = <rtcg.RTCFilterFunc> sample_wedge
        elif self.vpe == 8:
            self.filter_func = <rtcg.RTCFilterFunc> sample_hex
        else:
            raise NotImplementedError("Sampler type not implemented.")

        rtcg.rtcSetIntersectionFilterFunction(scene.scene_i,
                                              self.mesh,
                                              self.filter_func)

    def __dealloc__(self):
        free(self.field_data)
        free(self.element_indices)
        free(self.vertices)
        free(self.indices)


cdef class QuadraticElementMesh:
    r'''

    This creates a mesh of quadratic patches corresponding to the faces
    of 2nd-order Lagrange elements for direct rendering via embree.
    Currently, this is implemented for 20-point hexahedral meshes only.

    Parameters
    ----------

    scene : EmbreeScene
        This is the scene to which the constructed patches will be
        added.
    vertices : a np.ndarray of floats.
        This specifies the x, y, and z coordinates of the vertices in
        the mesh. This should either have the shape
        (num_vertices, 3). For example, vertices[2][1] should give the
        y-coordinate of the 3rd vertex in the mesh.
    indices : a np.ndarray of ints
        This should have the shape (num_elements, 20). Each hex will be
        represented in the scene by 6 bi-quadratic patches. We assume that
        the node ordering is as defined here:
        http://homepages.cae.wisc.edu/~tautges/papers/cnmev3.pdf

    '''

    cdef void* patches
    cdef np.float64_t* vertices
    cdef np.float64_t* field_data
    cdef unsigned int mesh
    # patches per element, vertices per element, vertices per face,
    # and field points per element, respectively:
    cdef int ppe, vpe, vpf, fpe

    def __init__(self, YTEmbreeScene scene,
                 np.ndarray vertices,
                 np.ndarray indices,
                 np.ndarray field_data):
        cdef int i, j

        # 20-point hexes
        if indices.shape[1] == 20:
            self.vpe = 20
            self.ppe = 6
            self.vpf = 8
            self._build_from_indices_hex20(scene, vertices, indices, field_data)
        # 10-point tets
        elif indices.shape[1] == 10:
            self.vpe = 10
            self.ppe = 4
            self.vpf = 6
            self._build_from_indices_tet10(scene, vertices, indices, field_data)
        else:
            raise NotImplementedError

    cdef void _build_from_indices_hex20(self, YTEmbreeScene scene,
                                  np.ndarray vertices_in,
                                  np.ndarray indices_in,
                                  np.ndarray field_data):
        cdef int i, j, k, ind, idim
        cdef int ne = indices_in.shape[0]
        cdef int nv = vertices_in.shape[0]
        cdef int npatch = self.ppe*ne;

        cdef unsigned int mesh = rtcgu.rtcNewUserGeometry(scene.scene_i, npatch)
        cdef np.ndarray[np.float64_t, ndim=2] element_vertices
        cdef Patch* patches = <Patch*> malloc(npatch * sizeof(Patch))
        self.vertices = <np.float64_t*> malloc(self.vpe * ne * 3 * sizeof(np.float64_t))
        self.field_data = <np.float64_t*> malloc(self.vpe * ne * sizeof(np.float64_t))

        for i in range(ne):
            element_vertices = vertices_in[indices_in[i]]
            for j in range(self.vpe):
                self.field_data[i*self.vpe + j] = field_data[i][j]
                for k in range(3):
                    self.vertices[i*self.vpe*3 + j*3 + k] = element_vertices[j][k]

        cdef Patch* patch
        for i in range(ne):  # for each element
            element_vertices = vertices_in[indices_in[i]]
            for j in range(self.ppe):  # for each face
                patch = &(patches[i*self.ppe+j])
                patch.geomID = mesh
                for k in range(self.vpf):  # for each vertex
                    ind = hex20_faces[j][k]
                    for idim in range(3):  # for each spatial dimension (yikes)
                        patch.v[k][idim] = element_vertices[ind][idim]
                patch.vertices = self.vertices + i*self.vpe*3
                patch.field_data = self.field_data + i*self.vpe

        self.patches = patches
        self.mesh = mesh

        rtcg.rtcSetUserData(scene.scene_i, self.mesh, patches)
        rtcgu.rtcSetBoundsFunction(scene.scene_i, self.mesh,
                                   <rtcgu.RTCBoundsFunc> patchBoundsFunc)
        rtcgu.rtcSetIntersectFunction(scene.scene_i, self.mesh,
                                      <rtcgu.RTCIntersectFunc> patchIntersectFunc)

    cdef void _build_from_indices_tet10(self, YTEmbreeScene scene,
                                  np.ndarray vertices_in,
                                  np.ndarray indices_in,
                                  np.ndarray field_data):
        cdef int i, j, k, ind, idim
        cdef int ne = indices_in.shape[0]
        cdef int nv = vertices_in.shape[0]
        cdef int npatch = self.ppe*ne;

        cdef unsigned int mesh = rtcgu.rtcNewUserGeometry(scene.scene_i, npatch)
        cdef np.ndarray[np.float64_t, ndim=2] element_vertices
        cdef Tet_Patch* patches = <Tet_Patch*> malloc(npatch * sizeof(Tet_Patch))
        self.vertices = <np.float64_t*> malloc(self.vpe * ne * 3 * sizeof(np.float64_t))
        self.field_data = <np.float64_t*> malloc(self.vpe * ne * sizeof(np.float64_t))

        for i in range(ne):
            element_vertices = vertices_in[indices_in[i]]
            for j in range(self.vpe):
                self.field_data[i*self.vpe + j] = field_data[i][j]
                for k in range(3):
                    self.vertices[i*self.vpe*3 + j*3 + k] = element_vertices[j][k]

        cdef Tet_Patch* patch
        for i in range(ne):  # for each element
            element_vertices = vertices_in[indices_in[i]]
            for j in range(self.ppe):  # for each face
                patch = &(patches[i*self.ppe+j])
                patch.geomID = mesh
                for k in range(self.vpf):  # for each vertex
                    ind = tet10_faces[j][k]
                    for idim in range(3):  # for each spatial dimension (yikes)
                        patch.v[k][idim] = element_vertices[ind][idim]
                patch.vertices = self.vertices + i*self.vpe*3
                patch.field_data = self.field_data + i*self.vpe

        self.patches = patches
        self.mesh = mesh

        rtcg.rtcSetUserData(scene.scene_i, self.mesh, patches)
        rtcgu.rtcSetBoundsFunction(scene.scene_i, self.mesh,
                                   <rtcgu.RTCBoundsFunc> tet_patchBoundsFunc)
        rtcgu.rtcSetIntersectFunction(scene.scene_i, self.mesh,
                                      <rtcgu.RTCIntersectFunc> tet_patchIntersectFunc)


    def __dealloc__(self):
        free(self.patches)
        free(self.vertices)
        free(self.field_data)



"""
This module gathers all user-facing functions with a `load_` prefix.

"""
import atexit
import os
import sys
import tarfile
import time
import types
import warnings
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import urlsplit

import numpy as np
from more_itertools import always_iterable

from yt.data_objects.static_output import Dataset
from yt.funcs import levenshtein_distance
from yt.sample_data.api import lookup_on_disk_data
from yt.utilities.decompose import decompose_array, get_psize
from yt.utilities.exceptions import (
    MountError,
    YTAmbiguousDataType,
    YTIllDefinedAMR,
    YTSimulationNotIdentified,
    YTUnidentifiedDataType,
)
from yt.utilities.hierarchy_inspection import find_lowest_subclasses
from yt.utilities.lib.misc_utilities import get_box_grids_level
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.object_registries import (
    output_type_registry,
    simulation_time_series_registry,
)
from yt.utilities.on_demand_imports import _pooch as pooch, _ratarmount as ratarmount

# --- Loaders for known data formats ---


def load(
    fn: Union[str, "os.PathLike[str]"], *args, hint: Optional[str] = None, **kwargs
):
    """
    Load a Dataset or DatasetSeries object.
    The data format is automatically discovered, and the exact return type is the
    corresponding subclass of :class:`yt.data_objects.static_output.Dataset`.
    A :class:`yt.data_objects.time_series.DatasetSeries` is created if the first
    argument is a pattern.

    Parameters
    ----------
    fn : str, os.Pathlike[str]
        A path to the data location. This can be a file name, directory name, a glob
        pattern, or a url (for data types that support it).

    hint : str, optional
        Only classes whose name include a hint are considered. If loading fails with
        a YTAmbiguousDataType exception, this argument can be used to lift ambiguity.
        Hints are case insensitive.

    Additional arguments, if any, are passed down to the return class.

    Returns
    -------
    :class:`yt.data_objects.static_output.Dataset` object
        If fn is a single path, create a Dataset from the appropriate subclass.

    :class:`yt.data_objects.time_series.DatasetSeries`
        If fn is a glob pattern (i.e. containing wildcards '[]?!*'), create a series.

    Raises
    ------
    FileNotFoundError
        If fn does not match any existing file or directory.

    yt.utilities.exceptions.YTUnidentifiedDataType
        If fn matches existing files or directories with undetermined format.

    yt.utilities.exceptions.YTAmbiguousDataType
        If the data format matches more than one class of similar specilization levels.
    """
    fn = os.fspath(fn)

    if any(wildcard in fn for wildcard in "[]?!*"):
        from yt.data_objects.time_series import DatasetSeries

        return DatasetSeries(fn, *args, hint=hint, **kwargs)

    # This will raise FileNotFoundError if the path isn't matched
    # either in the current dir or yt.config.ytcfg['data_dir_directory']
    if not fn.startswith("http"):
        fn = str(lookup_on_disk_data(fn))

    candidates = []
    for cls in output_type_registry.values():
        if cls._is_valid(fn, *args, **kwargs):
            candidates.append(cls)

    # Find only the lowest subclasses, i.e. most specialised front ends
    candidates = find_lowest_subclasses(candidates, hint=hint)

    if len(candidates) == 1:
        return candidates[0](fn, *args, **kwargs)

    if len(candidates) > 1:
        raise YTAmbiguousDataType(fn, candidates)

    raise YTUnidentifiedDataType(fn, *args, **kwargs)


def load_simulation(fn, simulation_type, find_outputs=False):
    """
    Load a simulation time series object of the specified simulation type.

    Parameters
    ----------
    fn : str, os.Pathlike, or byte (types supported by os.path.expandusers)
        Name of the data file or directory.

    simulation_type : str
        E.g. 'Enzo'

    find_outputs : bool
        Defaults to False

    Raises
    ------
    FileNotFoundError
        If fn is not found.

    yt.utilities.exceptions.YTSimulationNotIdentified
        If simulation_type is unknown.
    """

    fn = str(lookup_on_disk_data(fn))

    try:
        cls = simulation_time_series_registry[simulation_type]
    except KeyError as e:
        raise YTSimulationNotIdentified(simulation_type) from e

    return cls(fn, find_outputs=find_outputs)


# --- Loaders for generic ("stream") data ---


def load_uniform_grid(
    data,
    domain_dimensions,
    length_unit=None,
    bbox=None,
    nprocs=1,
    sim_time=0.0,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(True, True, True),
    geometry="cartesian",
    unit_system="cgs",
    default_species_fields=None,
    *,
    cell_widths=None,
    parameters=None,
):
    r"""Load a uniform grid of data into yt as a
    :class:`~yt.frontends.stream.data_structures.StreamHandler`.

    This should allow a uniform grid of data to be loaded directly into yt and
    analyzed as would any others.  This comes with several caveats:

    * Units will be incorrect unless the unit system is explicitly
      specified.
    * Some functions may behave oddly, and parallelism will be
      disappointing or non-existent in most cases.
    * Particles may be difficult to integrate.

    Particle fields are detected as one-dimensional fields.

    Parameters
    ----------
    data : dict
        This is a dict of numpy arrays, (numpy array, unit spec) tuples.
        Functions may also be supplied in place of numpy arrays as long as the
        subsequent argument nprocs is not specified to be greater than 1.
        Supplied functions much accepts the arguments (grid_object, field_name)
        and return numpy arrays.  The keys to the dict are the field names.
    domain_dimensions : array_like
        This is the domain dimensions of the grid
    length_unit : string
        Unit to use for lengths.  Defaults to unitless.
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units specified by length_unit.
        Defaults to a cubic unit-length domain.
    nprocs: integer, optional
        If greater than 1, will create this number of subarrays out of data
    sim_time : float, optional
        The simulation time in seconds
    mass_unit : string
        Unit to use for masses.  Defaults to unitless.
    time_unit : string
        Unit to use for times.  Defaults to unitless.
    velocity_unit : string
        Unit to use for velocities.  Defaults to unitless.
    magnetic_unit : string
        Unit to use for magnetic fields. Defaults to unitless.
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    geometry : string or tuple
        "cartesian", "cylindrical", "polar", "spherical", "geographic" or
        "spectral_cube".  Optionally, a tuple can be provided to specify the
        axis ordering -- for instance, to specify that the axis ordering should
        be z, x, y, this would be: ("cartesian", ("z", "x", "y")).  The same
        can be done for other coordinates, for instance:
        ("spherical", ("theta", "phi", "r")).
    default_species_fields : string, optional
        If set, default species fields are created for H and He which also
        determine the mean molecular weight. Options are "ionized" and "neutral".
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.

    Examples
    --------
    >>> np.random.seed(int(0x4D3D3D3))
    >>> bbox = np.array([[0.0, 1.0], [-1.5, 1.5], [1.0, 2.5]])
    >>> arr = np.random.random((128, 128, 128))
    >>> data = dict(density=arr)
    >>> ds = load_uniform_grid(data, arr.shape, length_unit="cm", bbox=bbox, nprocs=12)
    >>> dd = ds.all_data()
    >>> dd[("gas", "density")]
    unyt_array([0.76017901, 0.96855994, 0.49205428, ..., 0.78798258,
                0.97569432, 0.99453904], 'g/cm**3')
    """
    from yt.frontends.stream.data_structures import (
        StreamDataset,
        StreamDictFieldHandler,
        StreamHandler,
    )
    from yt.frontends.stream.definitions import (
        assign_particle_data,
        process_data,
        set_particle_types,
    )

    domain_dimensions = np.array(domain_dimensions)
    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], "float64")
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros(nprocs, dtype="int32").reshape((nprocs, 1))
    # First we fix our field names, apply units to data
    # and check for consistency of field shapes
    field_units, data, number_of_particles = process_data(
        data, grid_dims=tuple(domain_dimensions), allow_callables=nprocs == 1
    )

    sfh = StreamDictFieldHandler()

    if number_of_particles > 0:
        particle_types = set_particle_types(data)
        # Used much further below.
        pdata = {"number_of_particles": number_of_particles}
        for key in list(data.keys()):
            if len(data[key].shape) == 1 or key[0] == "io":
                if not isinstance(key, tuple):
                    field = ("io", key)
                    mylog.debug("Reassigning '%s' to '%s'", key, field)
                else:
                    field = key
                sfh._additional_fields += (field,)
                pdata[field] = data.pop(key)
    else:
        particle_types = {}

    if nprocs > 1:
        temp = {}
        new_data = {}
        for key in data.keys():
            psize = get_psize(np.array(data[key].shape), nprocs)
            grid_left_edges, grid_right_edges, shapes, slices = decompose_array(
                data[key].shape, psize, bbox
            )
            grid_dimensions = np.array([shape for shape in shapes], dtype="int32")
            temp[key] = [data[key][slice] for slice in slices]
        for gid in range(nprocs):
            new_data[gid] = {}
            for key in temp.keys():
                new_data[gid].update({key: temp[key][gid]})
        sfh.update(new_data)
        del new_data, temp
    else:
        sfh.update({0: data})
        grid_left_edges = domain_left_edge
        grid_right_edges = domain_right_edge
        grid_dimensions = domain_dimensions.reshape(nprocs, 3).astype("int32")

    if cell_widths is not None:
        # make sure this is a list, or else leave it as an empty guard value
        cell_widths = [cell_widths]

    if length_unit is None:
        length_unit = "code_length"
    if mass_unit is None:
        mass_unit = "code_mass"
    if time_unit is None:
        time_unit = "code_time"
    if velocity_unit is None:
        velocity_unit = "code_velocity"
    if magnetic_unit is None:
        magnetic_unit = "code_magnetic"

    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        -np.ones(nprocs, dtype="int64"),
        np.zeros(nprocs, dtype="int64").reshape(nprocs, 1),  # particle count
        np.zeros(nprocs).reshape((nprocs, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        cell_widths=cell_widths,
        parameters=parameters,
    )

    handler.name = "UniformGridData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = 2
    if np.all(domain_dimensions[1:] == 1):
        dimensionality = 1
    elif domain_dimensions[2] == 1:
        dimensionality = 2
    else:
        dimensionality = 3
    handler.dimensionality = dimensionality
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamDataset(
        handler,
        geometry=geometry,
        unit_system=unit_system,
        default_species_fields=default_species_fields,
    )

    # Now figure out where the particles go
    if number_of_particles > 0:
        # This will update the stream handler too
        assign_particle_data(sds, pdata, bbox)

    return sds


def load_amr_grids(
    grid_data,
    domain_dimensions,
    bbox=None,
    sim_time=0.0,
    length_unit=None,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(True, True, True),
    geometry="cartesian",
    refine_by=2,
    unit_system="cgs",
    default_species_fields=None,
    *,
    parameters=None,
):
    r"""Load a set of grids of data into yt as a
    :class:`~yt.frontends.stream.data_structures.StreamHandler`.
    This should allow a sequence of grids of varying resolution of data to be
    loaded directly into yt and analyzed as would any others.  This comes with
    several caveats:

    * Units will be incorrect unless the unit system is explicitly specified.
    * Some functions may behave oddly, and parallelism will be
      disappointing or non-existent in most cases.
    * Particles may be difficult to integrate.
    * No consistency checks are performed on the index

    Parameters
    ----------

    grid_data : list of dicts
        This is a list of dicts. Each dict must have entries "left_edge",
        "right_edge", "dimensions", "level", and then any remaining entries are
        assumed to be fields. Field entries must map to an NDArray *or* a
        function with the signature (grid_object, field_name) -> NDArray. The
        grid_data may also include a particle count. If no particle count is
        supplied, the dataset is understood to contain no particles. The
        grid_data will be modified in place and can't be assumed to be static.
        Grid data may also be supplied as a tuple of (NDarray or function, unit
        string) to specify the units.
    domain_dimensions : array_like
        This is the domain dimensions of the grid
    length_unit : string or float
        Unit to use for lengths.  Defaults to unitless.  If set to be a string, the bbox
        dimensions are assumed to be in the corresponding units.  If set to a float, the
        value is a assumed to be the conversion from bbox dimensions to centimeters.
    mass_unit : string or float
        Unit to use for masses.  Defaults to unitless.
    time_unit : string or float
        Unit to use for times.  Defaults to unitless.
    velocity_unit : string or float
        Unit to use for velocities.  Defaults to unitless.
    magnetic_unit : string or float
        Unit to use for magnetic fields.  Defaults to unitless.
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units specified by length_unit.
        Defaults to a cubic unit-length domain.
    sim_time : float, optional
        The simulation time in seconds
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    geometry : string or tuple
        "cartesian", "cylindrical", "polar", "spherical", "geographic" or
        "spectral_cube".  Optionally, a tuple can be provided to specify the
        axis ordering -- for instance, to specify that the axis ordering should
        be z, x, y, this would be: ("cartesian", ("z", "x", "y")).  The same
        can be done for other coordinates, for instance:
        ("spherical", ("theta", "phi", "r")).
    refine_by : integer or list/array of integers.
        Specifies the refinement ratio between levels.  Defaults to 2.  This
        can be an array, in which case it specifies for each dimension.  For
        instance, this can be used to say that some datasets have refinement of
        1 in one dimension, indicating that they span the full range in that
        dimension.
    default_species_fields : string, optional
        If set, default species fields are created for H and He which also
        determine the mean molecular weight. Options are "ionized" and "neutral".
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.

    Examples
    --------

    >>> grid_data = [
    ...     dict(
    ...         left_edge=[0.0, 0.0, 0.0],
    ...         right_edge=[1.0, 1.0, 1.0],
    ...         level=0,
    ...         dimensions=[32, 32, 32],
    ...         number_of_particles=0,
    ...     ),
    ...     dict(
    ...         left_edge=[0.25, 0.25, 0.25],
    ...         right_edge=[0.75, 0.75, 0.75],
    ...         level=1,
    ...         dimensions=[32, 32, 32],
    ...         number_of_particles=0,
    ...     ),
    ... ]
    ...
    >>> for g in grid_data:
    ...     g[("gas", "density")] = (
    ...         np.random.random(g["dimensions"]) * 2 ** g["level"],
    ...         "g/cm**3",
    ...     )
    ...
    >>> ds = load_amr_grids(grid_data, [32, 32, 32], length_unit=1.0)
    """
    from yt.frontends.stream.data_structures import (
        StreamDataset,
        StreamDictFieldHandler,
        StreamHandler,
    )
    from yt.frontends.stream.definitions import process_data, set_particle_types

    domain_dimensions = np.array(domain_dimensions)
    ngrids = len(grid_data)
    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], "float64")
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros((ngrids, 1), dtype="int32")
    grid_left_edges = np.zeros((ngrids, 3), dtype="float64")
    grid_right_edges = np.zeros((ngrids, 3), dtype="float64")
    grid_dimensions = np.zeros((ngrids, 3), dtype="int32")
    number_of_particles = np.zeros((ngrids, 1), dtype="int64")
    parent_ids = np.zeros(ngrids, dtype="int64") - 1
    sfh = StreamDictFieldHandler()
    for i, g in enumerate(grid_data):
        grid_left_edges[i, :] = g.pop("left_edge")
        grid_right_edges[i, :] = g.pop("right_edge")
        grid_dimensions[i, :] = g.pop("dimensions")
        grid_levels[i, :] = g.pop("level")
        field_units, data, n_particles = process_data(
            g, grid_dims=tuple(grid_dimensions[i, :])
        )
        number_of_particles[i, :] = n_particles
        sfh[i] = data

    # We now reconstruct our parent ids, so that our particle assignment can
    # proceed.
    mask = np.empty(ngrids, dtype="int32")
    for gi in range(ngrids):
        get_box_grids_level(
            grid_left_edges[gi, :],
            grid_right_edges[gi, :],
            grid_levels[gi] + 1,
            grid_left_edges,
            grid_right_edges,
            grid_levels,
            mask,
        )
        ids = np.where(mask.astype("bool"))
        for ci in ids:
            parent_ids[ci] = gi

    # Check if the grid structure is properly aligned (bug #1295)
    for lvl in range(grid_levels.min() + 1, grid_levels.max() + 1):
        idx = grid_levels.flatten() == lvl
        dims = domain_dimensions * refine_by ** (lvl - 1)
        for iax, ax in enumerate("xyz"):
            cell_edges = np.linspace(
                domain_left_edge[iax], domain_right_edge[iax], dims[iax], endpoint=False
            )
            if set(grid_left_edges[idx, iax]) - set(cell_edges):
                raise YTIllDefinedAMR(lvl, ax)

    if length_unit is None:
        length_unit = "code_length"
    if mass_unit is None:
        mass_unit = "code_mass"
    if time_unit is None:
        time_unit = "code_time"
    if velocity_unit is None:
        velocity_unit = "code_velocity"
    if magnetic_unit is None:
        magnetic_unit = "code_magnetic"

    particle_types = {}

    for grid in sfh.values():
        particle_types.update(set_particle_types(grid))

    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        parent_ids,
        number_of_particles,
        np.zeros(ngrids).reshape((ngrids, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        parameters=parameters,
    )

    handler.name = "AMRGridData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = refine_by
    if np.all(domain_dimensions[1:] == 1):
        dimensionality = 1
    elif domain_dimensions[2] == 1:
        dimensionality = 2
    else:
        dimensionality = 3
    handler.dimensionality = dimensionality
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamDataset(
        handler,
        geometry=geometry,
        unit_system=unit_system,
        default_species_fields=default_species_fields,
    )
    return sds


def load_particles(
    data,
    length_unit=None,
    bbox=None,
    sim_time=None,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(True, True, True),
    geometry="cartesian",
    unit_system="cgs",
    data_source=None,
    default_species_fields=None,
    *,
    parameters=None,
):
    r"""Load a set of particles into yt as a
    :class:`~yt.frontends.stream.data_structures.StreamParticleHandler`.

    This will allow a collection of particle data to be loaded directly into
    yt and analyzed as would any others.  This comes with several caveats:

    * There must be sufficient space in memory to contain all the particle
      data.
    * Parallelism will be disappointing or non-existent in most cases.
    * Fluid fields are not supported.

    Note: in order for the dataset to take advantage of SPH functionality,
    the following two fields must be provided:
    * ('io', 'density')
    * ('io', 'smoothing_length')

    Parameters
    ----------
    data : dict
        This is a dict of numpy arrays or (numpy array, unit name) tuples,
        where the keys are the field names. Particles positions must be named
        "particle_position_x", "particle_position_y", and "particle_position_z".
    length_unit : float
        Conversion factor from simulation length units to centimeters
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units of the length_unit
    sim_time : float, optional
        The simulation time in seconds
    mass_unit : float
        Conversion factor from simulation mass units to grams
    time_unit : float
        Conversion factor from simulation time units to seconds
    velocity_unit : float
        Conversion factor from simulation velocity units to cm/s
    magnetic_unit : float
        Conversion factor from simulation magnetic units to gauss
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    data_source : YTSelectionContainer, optional
        If set, parameters like `bbox`, `sim_time`, and code units are derived
        from it.
    default_species_fields : string, optional
        If set, default species fields are created for H and He which also
        determine the mean molecular weight. Options are "ionized" and "neutral".
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.

    Examples
    --------

    >>> pos = [np.random.random(128 * 128 * 128) for i in range(3)]
    >>> data = dict(
    ...     particle_position_x=pos[0],
    ...     particle_position_y=pos[1],
    ...     particle_position_z=pos[2],
    ... )
    >>> bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    >>> ds = load_particles(data, 3.08e24, bbox=bbox)

    """
    from yt.frontends.stream.data_structures import (
        StreamDictFieldHandler,
        StreamHandler,
        StreamParticlesDataset,
    )
    from yt.frontends.stream.definitions import process_data, set_particle_types

    domain_dimensions = np.ones(3, "int32")
    nprocs = 1

    # Parse bounding box
    if data_source is not None:
        le, re = data_source.get_bbox()
        le = le.to_value("code_length")
        re = re.to_value("code_length")
        bbox = list(zip(le, re))
    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], "float64")
    else:
        bbox = np.array(bbox)
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros(nprocs, dtype="int32").reshape((nprocs, 1))

    # Parse simulation time
    if data_source is not None:
        sim_time = data_source.ds.current_time
    if sim_time is None:
        sim_time = 0.0
    else:
        sim_time = float(sim_time)

    # Parse units
    def parse_unit(unit, dimension):
        if unit is None:
            unit = "code_" + dimension
            if data_source is not None:
                unit = getattr(data_source.ds, dimension + "_unit", unit)
        return unit

    length_unit = parse_unit(length_unit, "length")
    mass_unit = parse_unit(mass_unit, "mass")
    time_unit = parse_unit(time_unit, "time")
    velocity_unit = parse_unit(velocity_unit, "velocity")
    magnetic_unit = parse_unit(magnetic_unit, "magnetic")

    # Preprocess data
    field_units, data, _ = process_data(data)
    sfh = StreamDictFieldHandler()

    pdata = {}
    for key in data.keys():
        if not isinstance(key, tuple):
            field = ("io", key)
            mylog.debug("Reassigning '%s' to '%s'", key, field)
        else:
            field = key
        pdata[field] = data[key]
        sfh._additional_fields += (field,)
    data = pdata  # Drop reference count
    particle_types = set_particle_types(data)
    sfh.update({"stream_file": data})
    grid_left_edges = domain_left_edge
    grid_right_edges = domain_right_edge
    grid_dimensions = domain_dimensions.reshape(nprocs, 3).astype("int32")

    # I'm not sure we need any of this.
    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        -np.ones(nprocs, dtype="int64"),
        np.zeros(nprocs, dtype="int64").reshape(nprocs, 1),  # Temporary
        np.zeros(nprocs).reshape((nprocs, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        parameters=parameters,
    )

    handler.name = "ParticleData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = 2
    handler.dimensionality = 3
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamParticlesDataset(
        handler,
        geometry=geometry,
        unit_system=unit_system,
        default_species_fields=default_species_fields,
    )

    return sds


def load_hexahedral_mesh(
    data,
    connectivity,
    coordinates,
    length_unit=None,
    bbox=None,
    sim_time=0.0,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(True, True, True),
    geometry="cartesian",
    unit_system="cgs",
    *,
    parameters=None,
):
    r"""Load a hexahedral mesh of data into yt as a
    :class:`~yt.frontends.stream.data_structures.StreamHandler`.

    This should allow a semistructured grid of data to be loaded directly into
    yt and analyzed as would any others.  This comes with several caveats:

    * Units will be incorrect unless the data has already been converted to
      cgs.
    * Some functions may behave oddly, and parallelism will be
      disappointing or non-existent in most cases.
    * Particles may be difficult to integrate.

    Particle fields are detected as one-dimensional fields. The number of particles
    is set by the "number_of_particles" key in data.

    Parameters
    ----------
    data : dict
        This is a dict of numpy arrays, where the keys are the field names.
        There must only be one. Note that the data in the numpy arrays should
        define the cell-averaged value for of the quantity in in the hexahedral
        cell.
    connectivity : array_like
        This should be of size (N,8) where N is the number of zones.
    coordinates : array_like
        This should be of size (M,3) where M is the number of vertices
        indicated in the connectivity matrix.
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units of the length unit.
    sim_time : float, optional
        The simulation time in seconds
    mass_unit : string
        Unit to use for masses.  Defaults to unitless.
    time_unit : string
        Unit to use for times.  Defaults to unitless.
    velocity_unit : string
        Unit to use for velocities.  Defaults to unitless.
    magnetic_unit : string
        Unit to use for magnetic fields. Defaults to unitless.
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    geometry : string or tuple
        "cartesian", "cylindrical", "polar", "spherical", "geographic" or
        "spectral_cube".  Optionally, a tuple can be provided to specify the
        axis ordering -- for instance, to specify that the axis ordering should
        be z, x, y, this would be: ("cartesian", ("z", "x", "y")).  The same
        can be done for other coordinates, for instance:
        ("spherical", ("theta", "phi", "r")).
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.
    """
    from yt.frontends.stream.data_structures import (
        StreamDictFieldHandler,
        StreamHandler,
        StreamHexahedralDataset,
    )
    from yt.frontends.stream.definitions import process_data, set_particle_types

    domain_dimensions = np.ones(3, "int32") * 2
    nprocs = 1
    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], "float64")
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros(nprocs, dtype="int32").reshape((nprocs, 1))

    field_units, data, _ = process_data(data)
    sfh = StreamDictFieldHandler()

    particle_types = set_particle_types(data)

    sfh.update({"connectivity": connectivity, "coordinates": coordinates, 0: data})
    # Simple check for axis length correctness
    if len(data) > 0:
        fn = sorted(data)[0]
        array_values = data[fn]
        if array_values.size != connectivity.shape[0]:
            mylog.error(
                "Dimensions of array must be one fewer than the coordinate set."
            )
            raise RuntimeError
    grid_left_edges = domain_left_edge
    grid_right_edges = domain_right_edge
    grid_dimensions = domain_dimensions.reshape(nprocs, 3).astype("int32")

    if length_unit is None:
        length_unit = "code_length"
    if mass_unit is None:
        mass_unit = "code_mass"
    if time_unit is None:
        time_unit = "code_time"
    if velocity_unit is None:
        velocity_unit = "code_velocity"
    if magnetic_unit is None:
        magnetic_unit = "code_magnetic"

    # I'm not sure we need any of this.
    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        -np.ones(nprocs, dtype="int64"),
        np.zeros(nprocs, dtype="int64").reshape(nprocs, 1),  # Temporary
        np.zeros(nprocs).reshape((nprocs, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        parameters=parameters,
    )

    handler.name = "HexahedralMeshData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = 2
    handler.dimensionality = 3
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamHexahedralDataset(handler, geometry=geometry, unit_system=unit_system)

    return sds


def load_octree(
    octree_mask,
    data,
    bbox=None,
    sim_time=0.0,
    length_unit=None,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(True, True, True),
    over_refine_factor=None,
    num_zones=2,
    partial_coverage=1,
    unit_system="cgs",
    default_species_fields=None,
    *,
    parameters=None,
):
    r"""Load an octree mask into yt.

    Octrees can be saved out by calling save_octree on an OctreeContainer.
    This enables them to be loaded back in.

    This will initialize an Octree of data.  Note that fluid fields will not
    work yet, or possibly ever.

    Parameters
    ----------
    octree_mask : np.ndarray[uint8_t]
        This is a depth-first refinement mask for an Octree.  It should be
        of size n_octs * 8 (but see note about the root oct below), where
        each item is 1 for an oct-cell being refined and 0 for it not being
        refined.  For num_zones != 2, the children count will
        still be 8, so there will still be n_octs * 8 entries. Note that if
        the root oct is not refined, there will be only one entry
        for the root, so the size of the mask will be (n_octs - 1)*8 + 1.
    data : dict
        A dictionary of 1D arrays.  Note that these must of the size of the
        number of "False" values in the ``octree_mask``.
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units of length
    sim_time : float, optional
        The simulation time in seconds
    length_unit : string
        Unit to use for lengths.  Defaults to unitless.
    mass_unit : string
        Unit to use for masses.  Defaults to unitless.
    time_unit : string
        Unit to use for times.  Defaults to unitless.
    velocity_unit : string
        Unit to use for velocities.  Defaults to unitless.
    magnetic_unit : string
        Unit to use for magnetic fields. Defaults to unitless.
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    partial_coverage : boolean
        Whether or not an oct can be refined cell-by-cell, or whether all
        8 get refined.
    default_species_fields : string, optional
        If set, default species fields are created for H and He which also
        determine the mean molecular weight. Options are "ionized" and "neutral".
    num_zones : int
        The number of zones along each dimension in an oct
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.

    Example
    -------

    >>> import numpy as np
    >>> oct_mask = np.zeros(25)
    ... oct_mask[[0,  5,  7, 16]] = 8
    >>> octree_mask = np.array(oct_mask, dtype=np.uint8)
    >>> quantities = {}
    >>> quantities["gas", "density"] = np.random.random((22, 1))
    >>> bbox = np.array([[-10.0, 10.0], [-10.0, 10.0], [-10.0, 10.0]])

    >>> ds = load_octree(
    ...     octree_mask=octree_mask,
    ...     data=quantities,
    ...     bbox=bbox,
    ...     num_zones=1,
    ...     partial_coverage=0,
    ... )

    """
    from yt.frontends.stream.data_structures import (
        StreamDictFieldHandler,
        StreamHandler,
        StreamOctreeDataset,
    )
    from yt.frontends.stream.definitions import process_data, set_particle_types

    if not isinstance(octree_mask, np.ndarray) or octree_mask.dtype != np.uint8:
        raise TypeError("octree_mask should be a Numpy array with type uint8")

    nz = num_zones
    # for compatibility
    if over_refine_factor is not None:
        nz = 1 << over_refine_factor
    domain_dimensions = np.array([nz, nz, nz])
    nprocs = 1
    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], "float64")
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros(nprocs, dtype="int32").reshape((nprocs, 1))

    field_units, data, _ = process_data(data)
    sfh = StreamDictFieldHandler()

    particle_types = set_particle_types(data)

    sfh.update({0: data})
    grid_left_edges = domain_left_edge
    grid_right_edges = domain_right_edge
    grid_dimensions = domain_dimensions.reshape(nprocs, 3).astype("int32")

    if length_unit is None:
        length_unit = "code_length"
    if mass_unit is None:
        mass_unit = "code_mass"
    if time_unit is None:
        time_unit = "code_time"
    if velocity_unit is None:
        velocity_unit = "code_velocity"
    if magnetic_unit is None:
        magnetic_unit = "code_magnetic"

    # I'm not sure we need any of this.
    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        -np.ones(nprocs, dtype="int64"),
        np.zeros(nprocs, dtype="int64").reshape(nprocs, 1),  # Temporary
        np.zeros(nprocs).reshape((nprocs, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        parameters=parameters,
    )

    handler.name = "OctreeData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = 2
    handler.dimensionality = 3
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamOctreeDataset(
        handler, unit_system=unit_system, default_species_fields=default_species_fields
    )
    sds.octree_mask = octree_mask
    sds.partial_coverage = partial_coverage
    sds.num_zones = num_zones

    return sds


def load_unstructured_mesh(
    connectivity,
    coordinates,
    node_data=None,
    elem_data=None,
    length_unit=None,
    bbox=None,
    sim_time=0.0,
    mass_unit=None,
    time_unit=None,
    velocity_unit=None,
    magnetic_unit=None,
    periodicity=(False, False, False),
    geometry="cartesian",
    unit_system="cgs",
    *,
    parameters=None,
):
    r"""Load an unstructured mesh of data into yt as a
    :class:`~yt.frontends.stream.data_structures.StreamHandler`.

    This should allow an unstructured mesh data to be loaded directly into
    yt and analyzed as would any others.  Not all functionality for
    visualization will be present, and some analysis functions may not yet have
    been implemented.

    Particle fields are detected as one-dimensional fields. The number of
    particles is set by the "number_of_particles" key in data.

    In the parameter descriptions below, a "vertex" is a 3D point in space, an
    "element" is a single polyhedron whose location is defined by a set of
    vertices, and a "mesh" is a set of polyhedral elements, each with the same
    number of vertices.

    Parameters
    ----------

    connectivity : list of array_like or array_like
        This should either be a single 2D array or list of 2D arrays.  If this
        is a list, each element in the list corresponds to the connectivity
        information for a distinct mesh. Each array can have different
        connectivity length and should be of shape (N,M) where N is the number
        of elements and M is the number of vertices per element.
    coordinates : array_like
        The 3D coordinates of mesh vertices. This should be of size (L, D) where
        L is the number of vertices and D is the number of coordinates per vertex
        (the spatial dimensions of the dataset). Currently this must be either 2 or 3.
        When loading more than one mesh, the data for each mesh should be concatenated
        into a single coordinates array.
    node_data : dict or list of dicts
        For a single mesh, a dict mapping field names to 2D numpy arrays,
        representing data defined at element vertices. For multiple meshes,
        this must be a list of dicts.  Note that these are not the values as a
        function of the coordinates, but of the connectivity.  Their shape
        should be the same as the connectivity.  This means that if the data is
        in the shape of the coordinates, you may need to reshape them using the
        `connectivity` array as an index.
    elem_data : dict or list of dicts
        For a single mesh, a dict mapping field names to 1D numpy arrays, where
        each array has a length equal to the number of elements. The data
        must be defined at the center of each mesh element and there must be
        only one data value for each element. For multiple meshes, this must be
        a list of dicts, with one dict for each mesh.
    bbox : array_like (xdim:zdim, LE:RE), optional
        Size of computational domain in units of the length unit.
    sim_time : float, optional
        The simulation time in seconds
    length_unit : string
        Unit to use for length.  Defaults to unitless.
    mass_unit : string
        Unit to use for masses.  Defaults to unitless.
    time_unit : string
        Unit to use for times.  Defaults to unitless.
    velocity_unit : string
        Unit to use for velocities.  Defaults to unitless.
    magnetic_unit : string
        Unit to use for magnetic fields. Defaults to unitless.
    periodicity : tuple of booleans
        Determines whether the data will be treated as periodic along
        each axis
    geometry : string or tuple
        "cartesian", "cylindrical", "polar", "spherical", "geographic" or
        "spectral_cube".  Optionally, a tuple can be provided to specify the
        axis ordering -- for instance, to specify that the axis ordering should
        be z, x, y, this would be: ("cartesian", ("z", "x", "y")).  The same
        can be done for other coordinates, for instance:
        ("spherical", ("theta", "phi", "r")).
    parameters: dictionary, optional
        Optional dictionary used to populate the dataset parameters, useful
        for storing dataset metadata.

    Examples
    --------

    Load a simple mesh consisting of two tets.

      >>> # Coordinates for vertices of two tetrahedra
      >>> coordinates = np.array(
      ...     [
      ...         [0.0, 0.0, 0.5],
      ...         [0.0, 1.0, 0.5],
      ...         [0.5, 1, 0.5],
      ...         [0.5, 0.5, 0.0],
      ...         [0.5, 0.5, 1.0],
      ...     ]
      ... )
      >>> # The indices in the coordinates array of mesh vertices.
      >>> # This mesh has two elements.
      >>> connectivity = np.array([[0, 1, 2, 4], [0, 1, 2, 3]])

      >>> # Field data defined at the centers of the two mesh elements.
      >>> elem_data = {("connect1", "elem_field"): np.array([1, 2])}

      >>> # Field data defined at node vertices
      >>> node_data = {
      ...     ("connect1", "node_field"): np.array(
      ...         [[0.0, 1.0, 2.0, 4.0], [0.0, 1.0, 2.0, 3.0]]
      ...     )
      ... }

      >>> ds = load_unstructured_mesh(
      ...     connectivity, coordinates, elem_data=elem_data, node_data=node_data
      ... )
    """
    from yt.frontends.exodus_ii.util import get_num_pseudo_dims
    from yt.frontends.stream.data_structures import (
        StreamDictFieldHandler,
        StreamHandler,
        StreamUnstructuredMeshDataset,
    )
    from yt.frontends.stream.definitions import process_data, set_particle_types

    dimensionality = coordinates.shape[1]
    domain_dimensions = np.ones(3, "int32") * 2
    nprocs = 1

    if elem_data is None and node_data is None:
        raise RuntimeError("No data supplied in load_unstructured_mesh.")

    connectivity = list(always_iterable(connectivity, base_type=np.ndarray))
    num_meshes = max(1, len(connectivity))

    elem_data = list(always_iterable(elem_data, base_type=dict)) or [{}] * num_meshes
    node_data = list(always_iterable(node_data, base_type=dict)) or [{}] * num_meshes

    data = [{} for i in range(num_meshes)]
    for elem_dict, data_dict in zip(elem_data, data):
        for field, values in elem_dict.items():
            data_dict[field] = values
    for node_dict, data_dict in zip(node_data, data):
        for field, values in node_dict.items():
            data_dict[field] = values

    if bbox is None:
        bbox = [
            [
                coordinates[:, i].min() - 0.1 * abs(coordinates[:, i].min()),
                coordinates[:, i].max() + 0.1 * abs(coordinates[:, i].max()),
            ]
            for i in range(dimensionality)
        ]

    if dimensionality < 3:
        bbox.append([0.0, 1.0])
    if dimensionality < 2:
        bbox.append([0.0, 1.0])

    # handle pseudo-dims here
    num_pseudo_dims = get_num_pseudo_dims(coordinates)
    dimensionality -= num_pseudo_dims
    for i in range(dimensionality, 3):
        bbox[i][0] = 0.0
        bbox[i][1] = 1.0

    bbox = np.array(bbox, dtype=np.float64)
    domain_left_edge = np.array(bbox[:, 0], "float64")
    domain_right_edge = np.array(bbox[:, 1], "float64")
    grid_levels = np.zeros(nprocs, dtype="int32").reshape((nprocs, 1))

    field_units = {}
    particle_types = {}
    sfh = StreamDictFieldHandler()

    sfh.update({"connectivity": connectivity, "coordinates": coordinates})
    for i, d in enumerate(data):
        _f_unit, _data, _ = process_data(d)
        field_units.update(_f_unit)
        sfh[i] = _data
        particle_types.update(set_particle_types(d))

    grid_left_edges = domain_left_edge
    grid_right_edges = domain_right_edge
    grid_dimensions = domain_dimensions.reshape(nprocs, 3).astype("int32")

    if length_unit is None:
        length_unit = "code_length"
    if mass_unit is None:
        mass_unit = "code_mass"
    if time_unit is None:
        time_unit = "code_time"
    if velocity_unit is None:
        velocity_unit = "code_velocity"
    if magnetic_unit is None:
        magnetic_unit = "code_magnetic"

    # I'm not sure we need any of this.
    handler = StreamHandler(
        grid_left_edges,
        grid_right_edges,
        grid_dimensions,
        grid_levels,
        -np.ones(nprocs, dtype="int64"),
        np.zeros(nprocs, dtype="int64").reshape(nprocs, 1),  # Temporary
        np.zeros(nprocs).reshape((nprocs, 1)),
        sfh,
        field_units,
        (length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit),
        particle_types=particle_types,
        periodicity=periodicity,
        parameters=parameters,
    )

    handler.name = "UnstructuredMeshData"
    handler.domain_left_edge = domain_left_edge
    handler.domain_right_edge = domain_right_edge
    handler.refine_by = 2
    handler.dimensionality = dimensionality
    handler.domain_dimensions = domain_dimensions
    handler.simulation_time = sim_time
    handler.cosmology_simulation = 0

    sds = StreamUnstructuredMeshDataset(
        handler, geometry=geometry, unit_system=unit_system
    )

    fluid_types = ["all"]
    for i in range(1, num_meshes + 1):
        fluid_types += ["connect%d" % i]
    sds.fluid_types = tuple(fluid_types)

    def flatten(l):
        return [item for sublist in l for item in sublist]

    sds._node_fields = flatten([[f[1] for f in m] for m in node_data if m])
    sds._elem_fields = flatten([[f[1] for f in m] for m in elem_data if m])
    sds.default_field = [f for f in sds.field_list if f[0] == "connect1"][-1]
    sds.default_fluid_type = sds.default_field[0]
    return sds


# --- Loader for yt sample datasets ---
def load_sample(
    fn: Optional[str] = None, *, progressbar: bool = True, timeout=None, **kwargs
):
    r"""
    Load sample data with yt.

    This is a simple wrapper around :func:`~yt.loaders.load` to include fetching
    data with pooch from remote source.

    The data registry table can be retrieved and visualized using
    :func:`~yt.sample_data.api.get_data_registry_table`.
    The `filename` column contains usable keys that can be passed
    as the first positional argument to load_sample.
    Some data samples contain series of datasets. It may be required to
    supply the relative path to a specific dataset.

    Parameters
    ----------

    fn: str
        The `filename` of the dataset to load, as defined in the data registry
        table.

    progressbar: bool
        display a progress bar (tqdm).

    timeout: float or int (optional)
        Maximal waiting time, in seconds, after which download is aborted.
        `None` means "no limit". This parameter is directly passed to down to
        requests.get via pooch.HTTPDownloader

    Notes
    -----

    - This function is experimental as of yt 4.0.0, do not rely on its exact behaviour.
    - Any additional keyword argument is passed down to :func:`~yt.loaders.load`.
    - In case of collision with predefined keyword arguments as set in
      the data registry, the ones passed to this function take priority.
    - Datasets with slashes '/' in their names can safely be used even on Windows.
      On the contrary, paths using backslashes '\' won't work outside of Windows, so
      it is recommended to favour the UNIX convention ('/') in scripts that are meant
      to be cross-platform.
    - This function requires pandas and pooch.
    - Corresponding sample data live at https://yt-project.org/data

    """

    if fn is None:
        print(
            "One can see which sample datasets are available at: https://yt-project.org/data\n"
            "or alternatively by running: yt.sample_data.api.get_data_registry_table()",
            file=sys.stderr,
        )
        return None

    from yt.sample_data.api import (
        _download_sample_data_file,
        _get_test_data_dir_path,
        get_data_registry_table,
    )

    pooch_logger = pooch.utils.get_logger()

    # normalize path for platform portability
    # for consistency with yt.load, we also convert to str explicitly,
    # which gives us support Path objects for free
    fn = str(fn).replace("/", os.path.sep)

    topdir, _, specific_file = fn.partition(os.path.sep)

    registry_table = get_data_registry_table()

    known_names: List[str] = registry_table.dropna()["filename"].to_list()
    if topdir not in known_names:
        msg = f"'{topdir}' is not an available dataset."
        lexical_distances: List[Tuple[str, int]] = [
            (name, levenshtein_distance(name, topdir)) for name in known_names
        ]
        suggestions: List[str] = [name for name, dist in lexical_distances if dist < 4]
        if len(suggestions) == 1:
            msg += f" Did you mean '{suggestions[0]}' ?"
        elif suggestions:
            msg += " Did you mean to type any of the following ?\n\n    "
            msg += "\n    ".join(f"'{_}'" for _ in suggestions)
        raise ValueError(msg)

    # PR 3089
    # note: in the future the registry table should be reindexed
    # so that the following line can be replaced with
    #
    # specs = registry_table.loc[fn]
    #
    # however we don't want to do it right now because the "filename" column is
    # currently incomplete
    specs = registry_table.query(f"`filename` == '{topdir}'").iloc[0]

    load_name = specific_file or specs["load_name"] or ""

    if not isinstance(specs["load_kwargs"], dict):
        raise ValueError(
            "The requested dataset seems to be improperly registered.\n"
            "Tip: the entry in yt/sample_data_registry.json may be inconsistent with "
            "https://github.com/yt-project/website/blob/master/data/datafiles.json\n"
            "Please report this to https://github.com/yt-project/yt/issues/new"
        )

    kwargs = {**specs["load_kwargs"], **kwargs}

    save_dir = _get_test_data_dir_path()

    data_path = save_dir.joinpath(fn)
    if save_dir.joinpath(topdir).exists():
        # if the data is already available locally, `load_sample`
        # only acts as a thin wrapper around `load`
        if load_name and os.sep not in fn:
            data_path = data_path.joinpath(load_name)
        mylog.info("Sample dataset found in '%s'", data_path)
        if timeout is not None:
            mylog.info("Ignoring the `timeout` keyword argument received.")
        return load(data_path, **kwargs)

    mylog.info("'%s' is not available locally. Looking up online.", fn)

    # effectively silence the pooch's logger and create our own log instead
    pooch_logger.setLevel(100)
    mylog.info("Downloading from %s", specs["url"])

    # downloading via a pooch.Pooch instance behind the scenes
    filename = urlsplit(specs["url"]).path.split("/")[-1]

    tmp_file = _download_sample_data_file(
        filename, progressbar=progressbar, timeout=timeout
    )

    # pooch has functionalities to unpack downloaded archive files,
    # but it needs to be told in advance that we are downloading a tarball.
    # Since that information is not necessarily trivial to guess from the filename,
    # we rely on the standard library to perform a conditional unpacking instead.
    if tarfile.is_tarfile(tmp_file):
        mylog.info("Untaring downloaded file to '%s'", save_dir)
        with tarfile.open(tmp_file) as fh:

            def is_within_directory(directory, target):

                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)

                prefix = os.path.commonprefix([abs_directory, abs_target])

                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(fh, save_dir)
        os.remove(tmp_file)
    else:
        os.replace(tmp_file, os.path.join(save_dir, fn))

    loadable_path = Path.joinpath(save_dir, fn)
    if load_name not in str(loadable_path):
        loadable_path = loadable_path.joinpath(load_name, specific_file)

    return load(loadable_path, **kwargs)


def _mount_helper(
    archive: str, mountPoint: str, ratarmount_kwa: Dict, conn: Connection
):
    try:
        fuseOperationsObject = ratarmount.TarMount(
            pathToMount=archive,
            mountPoint=mountPoint,
            lazyMounting=True,
            **ratarmount_kwa,
        )
        fuseOperationsObject.use_ns = True
        conn.send(True)
    except Exception:
        conn.send(False)
        raise

    ratarmount.fuse.FUSE(
        operations=fuseOperationsObject,
        mountpoint=mountPoint,
        foreground=True,
        nothreads=True,
    )


# --- Loader for tar-based datasets ---
def load_archive(
    fn: Union[str, Path],
    path: str,
    ratarmount_kwa: Optional[Dict] = None,
    mount_timeout: float = 1.0,
    *args,
    **kwargs,
) -> Dataset:
    r"""
    Load archived data with yt.

    This is a wrapper around :func:`~yt.loaders.load` to include mounting
    and unmounting the archive as a read-only filesystem and load it.

    Parameters
    ----------

    fn: str
        The `filename` of the archive containing the dataset.

    path: str
        The path to the dataset in the archive.

    ratarmount_kwa: dict, optional
        Optional parameters to pass to ratarmount to mount the archive.

    mount_timeout: float, optional
        The timeout to wait for ratarmount to mount the archive. Default is 1s.

    Notes
    -----

    - The function is experimental and may work or not depending on your setup.
    - Any additional keyword argument is passed down to :func:`~yt.loaders.load`.
    - This function requires ratarmount to be installed.
    - This function does not work on Windows system.
    """

    warnings.warn(
        "The 'load_archive' function is still experimental and may be unstable."
    )

    fn = os.path.expanduser(fn)

    # This will raise FileNotFoundError if the path isn't matched
    # either in the current dir or yt.config.ytcfg['data_dir_directory']
    if not fn.startswith("http"):
        fn = str(lookup_on_disk_data(fn))

    if ratarmount_kwa is None:
        ratarmount_kwa = {}

    try:
        tarfile.open(fn)
    except tarfile.ReadError:
        raise YTUnidentifiedDataType(fn, *args, **kwargs)

    # Note: the temporary directory will be created by ratarmount
    tempdir = fn + ".mount"
    tempdir_base = tempdir
    i = 0
    while os.path.exists(tempdir):
        i += 1
        tempdir = f"{tempdir_base}.{i}"

    parent_conn, child_conn = Pipe()
    proc = Process(target=_mount_helper, args=(fn, tempdir, ratarmount_kwa, child_conn))
    proc.start()
    if not parent_conn.recv():
        raise MountError(f"An error occured while mounting {fn} in {tempdir}")

    # Note: the mounting needs to happen in another process which
    # needs be run in the foreground (otherwise it may
    # unmount). To prevent a race-condition here, we wait
    # for the folder to be mounted within a reasonable time.
    t = 0.0
    while t < mount_timeout:
        if os.path.ismount(tempdir):
            break
        time.sleep(0.1)
        t += 0.1
    else:
        raise MountError(f"Folder {tempdir} does not appear to be mounted")

    # We need to kill the process at exit (to force unmounting)
    def umount_callback():
        proc.terminate()

    atexit.register(umount_callback)

    # Alternatively, can dismount manually
    def del_callback(self):
        proc.terminate()
        atexit.unregister(umount_callback)

    ds = load(os.path.join(tempdir, path), *args, **kwargs)
    ds.dismount = types.MethodType(del_callback, ds)

    return ds


def load_hdf5_file(
    fn: Union[str, "os.PathLike[str]"],
    root_node: Optional[str] = "/",
    fields: Optional[List[str]] = None,
    bbox: np.ndarray = None,
    nchunks: int = 0,
    dataset_arguments: Optional[dict] = None,
):
    """
    Create a (grid-based) yt dataset given the path to an hdf5 file.

    This function accepts a filename, as well as (potentially) a bounding box,
    the root node where fields are stored, and the number of chunks to attempt
    to decompose the object into.  This function will then introspect that HDF5
    file, attempt to determine the available fields, and then supply a
    :class:`yt.data_objects.static_output.Dataset` object.  However, unlike the
    other loaders, the data is *not* required to be preloaded into memory, and will
    only be loaded *on demand*.

    This does not yet work with particle-type datasets.

    Parameters
    ----------
    fn : str, os.Pathlike[str]
        A path to the hdf5 file that contains the data.

    root_node: str, optional
        If the fields to be loaded are stored under an HDF5 group object,
        specify it here.  Otherwise, the fields are assumed to be at the root
        level of the HDF5 file hierarchy.

    fields : list of str, optional
        The fields to be included as part of the dataset.  If this is not
        included, all of the datasets under *root_node* will be included.
        If your file contains, for instance, a "parameters" node at the root
        level next to other fields, it would be (mistakenly) included.  This
        allows you to specify only those that should be included.

    bbox : array_like (xdim:zdim, LE:RE), optional
        If supplied, this will be the bounding box for the dataset.  If not
        supplied, it will be assumed to be from 0 to 1 in all dimensions.

    nchunks : int, optional
        How many chunks should this dataset be split into?  If 0 or not
        supplied, yt will attempt to ensure that there is one chunk for every
        64**3 zones in the dataset.

    dataset_arguments : dict, optional
        Any additional arguments that should be passed to
        :class:`yt.loaders.load_amr_grids`, including things like the unit
        length and the coordinates.

    Returns
    -------
    :class:`yt.data_objects.static_output.Dataset` object
        This returns an instance of a dataset, created with `load_amr_grids`,
        that can read from the HDF5 file supplied to this function.  An open
        handle to the HDF5 file is retained.

    Raises
    ------
    FileNotFoundError
        If fn does not match any existing file or directory.

    """

    from yt.utilities.on_demand_imports import _h5py as h5py

    dataset_arguments = dataset_arguments or {}

    if bbox is None:
        bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
        mylog.info("Assuming unitary (0..1) bounding box.")

    def _read_data(handle, root_node):
        def _reader(grid, field_name):
            ftype, fname = field_name
            si = grid.get_global_startindex()
            ei = si + grid.ActiveDimensions
            return handle[root_node][fname][si[0] : ei[0], si[1] : ei[1], si[2] : ei[2]]

        return _reader

    fn = str(lookup_on_disk_data(fn))
    handle = h5py.File(fn, "r")
    reader = _read_data(handle, root_node)
    if fields is None:
        fields = list(handle[root_node].keys())
        mylog.debug("Identified fields %s", fields)
    shape = handle[root_node][fields[0]].shape
    if nchunks <= 0:
        # We apply a pretty simple heuristic here.  We don't want more than
        # about 64^3 zones per chunk.  So ...
        full_size = np.prod(shape)
        nchunks = full_size // (64**3)
        mylog.info("Auto-guessing %s chunks from a size of %s", nchunks, full_size)
    grid_data = []
    psize = get_psize(np.array(shape), nchunks)
    left_edges, right_edges, shapes, _ = decompose_array(shape, psize, bbox)
    for le, re, s in zip(left_edges, right_edges, shapes):
        data = {_: reader for _ in fields}
        data.update({"left_edge": le, "right_edge": re, "dimensions": s, "level": 0})
        grid_data.append(data)
    return load_amr_grids(grid_data, shape, **dataset_arguments)



from typing import Type

import numpy as np

import yt.units
from yt.fields.field_info_container import FieldInfoContainer
from yt.frontends.gadget.data_structures import GadgetHDF5Dataset
from yt.frontends.owls.fields import OWLSFieldInfo
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import EagleNetworkFieldInfo


class EagleDataset(GadgetHDF5Dataset):
    _particle_mass_name = "Mass"
    _field_info_class: Type[FieldInfoContainer] = OWLSFieldInfo
    _time_readin_ = "Time"

    def _parse_parameter_file(self):

        # read values from header
        hvals = self._get_hvals()
        self.parameters = hvals

        # set features common to OWLS and Eagle
        self._set_owls_eagle()

        # Set time from analytic solution for flat LCDM universe
        a = hvals["ExpansionFactor"]
        H0 = hvals["H(z)"] / hvals["E(z)"]
        a_eq = (self.omega_matter / self.omega_lambda) ** (1.0 / 3)
        t1 = 2.0 / (3.0 * np.sqrt(self.omega_lambda))
        t2 = (a / a_eq) ** (3.0 / 2)
        t3 = np.sqrt(1.0 + (a / a_eq) ** 3)
        t = t1 * np.log(t2 + t3) / H0
        self.current_time = t * yt.units.s

    def _set_code_unit_attributes(self):
        self._set_owls_eagle_units()

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        need_groups = [
            "Config",
            "Constants",
            "HashTable",
            "Header",
            "Parameters",
            "RuntimePars",
            "Units",
        ]
        veto_groups = [
            "SUBFIND",
            "PartType0/ChemistryAbundances",
            "PartType0/ChemicalAbundances",
        ]
        valid = True
        try:
            fileh = h5py.File(filename, mode="r")
            for ng in need_groups:
                if ng not in fileh["/"]:
                    valid = False
            for vg in veto_groups:
                if vg in fileh["/"]:
                    valid = False
            fileh.close()
        except Exception:
            valid = False
            pass
        return valid


class EagleNetworkDataset(EagleDataset):
    _particle_mass_name = "Mass"
    _field_info_class = EagleNetworkFieldInfo
    _time_readin = "Time"

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        try:
            fileh = h5py.File(filename, mode="r")
            if (
                "Constants" in fileh["/"].keys()
                and "Header" in fileh["/"].keys()
                and "SUBFIND" not in fileh["/"].keys()
                and (
                    "ChemistryAbundances" in fileh["PartType0"].keys()
                    or "ChemicalAbundances" in fileh["PartType0"].keys()
                )
            ):
                fileh.close()
                return True
            fileh.close()
        except Exception:
            pass
        return False



import yt

# Load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a projection and save it with the default colormap ('cmyt.arbre')
p = yt.ProjectionPlot(ds, "z", ("gas", "density"), width=(100, "kpc"))
p.save()

# Change the colormap to 'cmyt.dusk' and save again.  We must specify
# a different filename here or it will save it over the top of
# our first projection.
p.set_cmap(field=("gas", "density"), cmap="cmyt.dusk")
p.save("proj_with_dusk_cmap.png")

# Change the colormap to 'hot' and save again.
p.set_cmap(field=("gas", "density"), cmap="hot")
p.save("proj_with_hot_cmap.png")



from yt.testing import assert_array_equal, assert_raises, fake_amr_ds, fake_random_ds
from yt.units import cm


def test_box_creation():

    # test that creating a region with left and right edge
    # with units works
    ds = fake_random_ds(32, length_unit=2)
    reg = ds.box([0, 0, 0] * cm, [2, 2, 2] * cm)
    dens_units = reg[("gas", "density")]

    reg = ds.box([0, 0, 0], [1, 1, 1])
    dens_no_units = reg[("gas", "density")]

    assert_array_equal(dens_units, dens_no_units)


def test_max_level_min_level_semantics():
    ds = fake_amr_ds()
    ad = ds.all_data()
    assert ad[("index", "grid_level")].max() == 4
    ad.max_level = 2
    assert ad[("index", "grid_level")].max() == 2
    ad.max_level = 8
    assert ad[("index", "grid_level")].max() == 4
    ad.min_level = 2
    assert ad[("index", "grid_level")].min() == 2
    ad.min_level = 0
    assert ad[("index", "grid_level")].min() == 0


def test_ellipsis_selection():
    ds = fake_amr_ds()
    reg = ds.r[:, :, :]
    ereg = ds.r[...]
    assert_array_equal(reg.fwidth, ereg.fwidth)

    reg = ds.r[(0.5, "cm"), :, :]
    ereg = ds.r[(0.5, "cm"), ...]
    assert_array_equal(reg.fwidth, ereg.fwidth)

    reg = ds.r[:, :, (0.5, "cm")]
    ereg = ds.r[..., (0.5, "cm")]
    assert_array_equal(reg.fwidth, ereg.fwidth)

    reg = ds.r[:, :, (0.5, "cm")]
    ereg = ds.r[..., (0.5, "cm")]
    assert_array_equal(reg.fwidth, ereg.fwidth)

    assert_raises(IndexError, ds.r.__getitem__, (..., (0.5, "cm"), ...))


# this test will fail until "arbitrary_grid" selector is implemented for 2D datasets
# see issue https://github.com/yt-project/yt/issues/3437
"""
from yt.utilities.answer_testing.framework import data_dir_load, requires_ds

@requires_ds("castro_sedov_2d_cyl_in_cart_plt00150")
def test_complex_slicing_2D_consistency():

    # see https://github.com/yt-project/yt/issues/3429
    ds = data_dir_load("castro_sedov_2d_cyl_in_cart_plt00150")

    reg = ds.r[0.1:0.2:8j, :]
    reg["gas", "density"]
    reg = ds.r[:, 1:2:8j]
    reg["gas", "density"]
    reg = ds.r[0.1:0.2:8j, 1:2:8j]
    reg["gas", "density"]
"""



import pytest

from yt.data_objects.static_output import Dataset
from yt.geometry.grid_geometry_handler import GridIndex
from yt.loaders import load, load_simulation
from yt.utilities.exceptions import (
    YTAmbiguousDataType,
    YTSimulationNotIdentified,
    YTUnidentifiedDataType,
)
from yt.utilities.object_registries import output_type_registry


@pytest.fixture
def tmp_data_dir(tmp_path):
    from yt.config import ytcfg

    pre_test_data_dir = ytcfg["yt", "test_data_dir"]
    ytcfg.set("yt", "test_data_dir", str(tmp_path))

    yield tmp_path

    ytcfg.set("yt", "test_data_dir", pre_test_data_dir)


@pytest.mark.usefixtures("tmp_data_dir")
def test_load_not_a_file(tmp_path):
    with pytest.raises(FileNotFoundError):
        load(tmp_path / "not_a_file")


@pytest.mark.parametrize("simtype", ["Enzo", "unregistered_simulation_type"])
@pytest.mark.usefixtures("tmp_data_dir")
def test_load_simulation_not_a_file(simtype, tmp_path):
    # it is preferable to report the most important problem in an error message
    # (missing data is worse than a typo insimulation_type)
    # so we make sure the error raised is not YTSimulationNotIdentified,
    # even with an absurd simulation type
    with pytest.raises(FileNotFoundError):
        load_simulation(tmp_path / "not_a_file", simtype)


@pytest.fixture()
def tmp_path_with_empty_file(tmp_path):
    empty_file_path = tmp_path / "empty_file"
    empty_file_path.touch()
    return tmp_path, empty_file_path


def test_load_unidentified_data_dir(tmp_path_with_empty_file):
    tmp_path, empty_file_path = tmp_path_with_empty_file
    with pytest.raises(YTUnidentifiedDataType):
        load(tmp_path)


def test_load_unidentified_data_file(tmp_path_with_empty_file):
    tmp_path, empty_file_path = tmp_path_with_empty_file
    with pytest.raises(YTUnidentifiedDataType):
        load(empty_file_path)


def test_load_simulation_unidentified_data_dir(tmp_path_with_empty_file):
    tmp_path, empty_file_path = tmp_path_with_empty_file
    with pytest.raises(YTSimulationNotIdentified):
        load_simulation(tmp_path, "unregistered_simulation_type")


def test_load_simulation_unidentified_data_file(tmp_path_with_empty_file):
    tmp_path, empty_file_path = tmp_path_with_empty_file
    with pytest.raises(YTSimulationNotIdentified):
        load_simulation(
            empty_file_path,
            "unregistered_simulation_type",
        )


@pytest.fixture()
def ambiguous_dataset_classes():
    # We deliberately setup a situation where two Dataset subclasses
    # that aren't parents are consisdered valid.
    # We implement the bare minimum for these classes to be actually
    # loadable in order to test hints.
    class MockHierarchy(GridIndex):
        pass

    class MockDataset(Dataset):
        _index_class = MockHierarchy

        def _parse_parameter_file(self, *args, **kwargs):
            self.current_time = -1.0
            self.cosmological_simulation = 0

        def _set_code_unit_attributes(self, *args, **kwargs):
            self.length_unit = self.quan(1, "m")
            self.mass_unit = self.quan(1, "kg")
            self.time_unit = self.quan(1, "s")

    class AlphaDataset(MockDataset):
        @classmethod
        def _is_valid(cls, *args, **kwargs):
            return True

    class BetaDataset(MockDataset):
        @classmethod
        def _is_valid(cls, *args, **kwargs):
            return True

    yield

    # teardown to avoid possible breakage in following tests
    output_type_registry.pop("MockDataset")
    output_type_registry.pop("AlphaDataset")
    output_type_registry.pop("BetaDataset")


@pytest.mark.usefixtures("ambiguous_dataset_classes")
def test_load_ambiguous_data(tmp_path):
    with pytest.raises(YTAmbiguousDataType):
        load(tmp_path)

    file = tmp_path / "fake_datafile0011.dump"
    file.touch()

    pattern = str(tmp_path / "fake_datafile00??.dump")

    # loading a DatasetSeries should not crash until an item is retrieved
    ts = load(pattern)
    with pytest.raises(YTAmbiguousDataType):
        ts[0]


@pytest.mark.parametrize(
    "hint, expected_type",
    [
        ("alpha", "AlphaDataset"),
        ("al", "AlphaDataset"),
        ("ph", "AlphaDataset"),
        ("beta", "BetaDataset"),
        ("BeTA", "BetaDataset"),
        ("b", "BetaDataset"),
    ],
)
@pytest.mark.usefixtures("ambiguous_dataset_classes")
def test_load_ambiguous_data_with_hint(hint, expected_type, tmp_path):
    ds = load(tmp_path, hint=hint)
    assert type(ds).__name__ == expected_type

    file1 = tmp_path / "fake_datafile0011.dump"
    file2 = tmp_path / "fake_datafile0022.dump"
    file1.touch()
    file2.touch()

    pattern = str(tmp_path / "fake_datafile00??.dump")

    ts = load(pattern, hint=hint)
    ds = ts[0]
    assert type(ds).__name__ == expected_type

    ds = ts[1]
    assert type(ds).__name__ == expected_type



import yt

# Load the dataset.
ds = yt.load("Enzo_64/DD0029/data0029")

# Create a 1 Mpc radius sphere, centered on the max density.
sp = ds.sphere("max", (1.0, "Mpc"))

# Use the total_quantity derived quantity to sum up the
# values of the mass and particle_mass fields
# within the sphere.
baryon_mass, particle_mass = sp.quantities.total_quantity(
    [("gas", "mass"), ("all", "particle_mass")]
)

print(
    "Total mass in sphere is %0.3e Msun (gas = %0.3e Msun, particles = %0.3e Msun)"
    % (
        (baryon_mass + particle_mass).in_units("Msun"),
        baryon_mass.in_units("Msun"),
        particle_mass.in_units("Msun"),
    )
)



import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid

import yt

fn = "IsolatedGalaxy/galaxy0030/galaxy0030"
ds = yt.load(fn)  # load data

fig = plt.figure()

# See http://matplotlib.org/mpl_toolkits/axes_grid/api/axes_grid_api.html
# These choices of keyword arguments produce two colorbars, both drawn on the
# right hand side.  This means there are only two colorbar axes, one for Density
# and another for temperature.  In addition, axes labels will be drawn for all
# plots.
grid = AxesGrid(
    fig,
    (0.075, 0.075, 0.85, 0.85),
    nrows_ncols=(2, 2),
    axes_pad=1.0,
    label_mode="all",
    share_all=True,
    cbar_location="right",
    cbar_mode="edge",
    cbar_size="5%",
    cbar_pad="0%",
)

cuts = ["x", "y", "z", "z"]
fields = [
    ("gas", "density"),
    ("gas", "density"),
    ("gas", "density"),
    ("gas", "temperature"),
]

for i, (direction, field) in enumerate(zip(cuts, fields)):
    # Load the data and create a single plot
    p = yt.SlicePlot(ds, direction, field)
    p.zoom(40)

    # This forces the ProjectionPlot to redraw itself on the AxesGrid axes.
    plot = p.plots[field]
    plot.figure = fig
    plot.axes = grid[i].axes

    # Since there are only two colorbar axes, we need to make sure we don't try
    # to set the temperature colorbar to cbar_axes[4], which would if we used i
    # to index cbar_axes, yielding a plot without a temperature colorbar.
    # This unnecessarily redraws the Density colorbar three times, but that has
    # no effect on the final plot.
    if field == ("gas", "density"):
        plot.cax = grid.cbar_axes[0]
    elif field == ("gas", "temperature"):
        plot.cax = grid.cbar_axes[1]

    # Finally, redraw the plot.
    p._setup_plots()

plt.savefig("multiplot_2x2_coordaxes_slice.png")



# distutils: libraries = STD_LIBS
# distutils: include_dirs = LIB_DIR
"""
Fast hashing routines


"""


import numpy as np

cimport cython
cimport numpy as np


@cython.wraparound(False)
@cython.boundscheck(False)
cdef np.int64_t c_fnv_hash(unsigned char[:] octets) nogil:
    # https://bitbucket.org/yt_analysis/yt/issues/1052/field-access-tests-fail-under-python3
    # FNV hash cf. http://www.isthe.com/chongo/tech/comp/fnv/index.html
    cdef np.int64_t hash_val = 2166136261
    cdef unsigned char octet
    cdef int i
    for i in range(octets.shape[0]):
        hash_val = hash_val ^ octets[i]
        hash_val = hash_val * 16777619
    return hash_val

def fnv_hash(octets):
    """

    Create a FNV hash from a bytestring.
    Info: http://www.isthe.com/chongo/tech/comp/fnv/index.html

    Parameters
    ----------
    octets : bytestring
        The string of bytes to generate a hash from.
    """
    return c_fnv_hash(octets)



from unyt.unit_object import Unit
from unyt.unit_systems import add_symbols

from yt.units.unit_registry import default_unit_registry

add_symbols(globals(), registry=default_unit_registry)


class _SymbolContainer:
    """A container for units to associate with a dataset.

    This object is usually accessed on a Dataset instance via
    ``ds.units.unit_symbols``.

    Parameters
    ----------
    registry : UnitRegistry instance
        A unit registry to associate with units accessed on this object.

    Example
    -------

    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> code_mass = ds.units.code_mass
    >>> (12 * code_mass).to("Msun")
    unyt_quantity(4.89719136e+11, 'Msun')
    >>> code_mass.registry is ds.unit_registry
    True
    """

    def __init__(self, registry):
        self._registry = registry
        self._cache = {}

    def __dir__(self):
        ret = [u for u in globals() if not u.startswith("_")]
        ret += list(self._registry.keys())
        ret += object.__dir__(self)
        return list(set(ret))

    def __getattr__(self, item):
        if item in self._cache:
            return self._cache[item]
        if hasattr(globals(), item):
            ret = Unit(globals()[item].expr, registry=self._registry)
        elif item in self._registry:
            ret = Unit(item, registry=self._registry)
        else:
            raise AttributeError(item)
        self._cache[item] = ret
        return ret



import numpy as np

from yt.utilities.lib.misc_utilities import gravitational_binding_energy
from yt.utilities.operator_registry import OperatorRegistry
from yt.utilities.physical_constants import gravitational_constant_cgs as G

clump_validator_registry = OperatorRegistry()


def add_validator(name, function):
    clump_validator_registry[name] = ClumpValidator(function)


class ClumpValidator:
    r"""
    A ClumpValidator is a function that takes a clump and returns
    True or False as to whether the clump is valid and shall be kept.
    """

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None:
            self.args = []
        self.kwargs = kwargs
        if self.kwargs is None:
            self.kwargs = {}

    def __call__(self, clump):
        return self.function(clump, *self.args, **self.kwargs)


def _gravitationally_bound(
    clump, use_thermal_energy=True, use_particles=True, truncate=True, num_threads=0
):
    "True if clump is gravitationally bound."

    use_particles &= ("all", "particle_mass") in clump.data.ds.field_info

    bulk_velocity = clump.quantities.bulk_velocity(use_particles=use_particles)

    kinetic = (
        0.5
        * (
            clump["gas", "mass"]
            * (
                (bulk_velocity[0] - clump["gas", "velocity_x"]) ** 2
                + (bulk_velocity[1] - clump["gas", "velocity_y"]) ** 2
                + (bulk_velocity[2] - clump["gas", "velocity_z"]) ** 2
            )
        ).sum()
    )

    if use_thermal_energy:
        kinetic += (
            clump["gas", "mass"] * clump["gas", "specific_thermal_energy"]
        ).sum()

    if use_particles:
        kinetic += (
            0.5
            * (
                clump["all", "particle_mass"]
                * (
                    (bulk_velocity[0] - clump["all", "particle_velocity_x"]) ** 2
                    + (bulk_velocity[1] - clump["all", "particle_velocity_y"]) ** 2
                    + (bulk_velocity[2] - clump["all", "particle_velocity_z"]) ** 2
                )
            ).sum()
        )

    if use_particles:
        m = np.concatenate(
            [clump["gas", "mass"].in_cgs(), clump["all", "particle_mass"].in_cgs()]
        )
        px = np.concatenate(
            [clump["index", "x"].in_cgs(), clump["all", "particle_position_x"].in_cgs()]
        )
        py = np.concatenate(
            [clump["index", "y"].in_cgs(), clump["all", "particle_position_y"].in_cgs()]
        )
        pz = np.concatenate(
            [clump["index", "z"].in_cgs(), clump["all", "particle_position_z"].in_cgs()]
        )
    else:
        m = clump["gas", "mass"].in_cgs()
        px = clump["index", "x"].in_cgs()
        py = clump["index", "y"].in_cgs()
        pz = clump["index", "z"].in_cgs()

    potential = clump.data.ds.quan(
        G
        * gravitational_binding_energy(
            m, px, py, pz, truncate, (kinetic / G).in_cgs(), num_threads=num_threads
        ),
        kinetic.in_cgs().units,
    )

    if truncate and potential >= kinetic:
        return True

    return potential >= kinetic


add_validator("gravitationally_bound", _gravitationally_bound)


def _min_cells(clump, n_cells):
    "True if clump has a minimum number of cells."
    return clump["index", "ones"].size >= n_cells


add_validator("min_cells", _min_cells)



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Particle smoothing in cells




"""


cimport numpy as np

import numpy as np

cimport cython
from cpython.exc cimport PyErr_CheckSignals
from libc.math cimport cos, fabs, sin, sqrt
from libc.stdlib cimport free, malloc, realloc
from libc.string cimport memmove
from oct_container cimport Oct, OctInfo, OctreeContainer


cdef void spherical_coord_setup(np.float64_t ipos[3], np.float64_t opos[3]):
    opos[0] = ipos[0] * sin(ipos[1]) * cos(ipos[2])
    opos[1] = ipos[0] * sin(ipos[1]) * sin(ipos[2])
    opos[2] = ipos[0] * cos(ipos[1])

cdef void cart_coord_setup(np.float64_t ipos[3], np.float64_t opos[3]):
    opos[0] = ipos[0]
    opos[1] = ipos[1]
    opos[2] = ipos[2]

cdef class ParticleSmoothOperation:
    def __init__(self, nvals, nfields, max_neighbors, kernel_name):
        # This is the set of cells, in grids, blocks or octs, we are handling.
        cdef int i
        self.nvals = nvals
        self.nfields = nfields
        self.maxn = max_neighbors
        self.sph_kernel = get_kernel_func(kernel_name)

    def initialize(self, *args):
        raise NotImplementedError

    def finalize(self, *args):
        raise NotImplementedError

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    def process_octree(self, OctreeContainer mesh_octree,
                     np.int64_t [:] mdom_ind,
                     np.float64_t[:,:] positions,
                     np.float64_t[:,:] oct_positions,
                     fields = None, int domain_id = -1,
                     int domain_offset = 0,
                     periodicity = (True, True, True),
                     index_fields = None,
                     OctreeContainer particle_octree = None,
                     np.int64_t [:] pdom_ind = None,
                     geometry = "cartesian"):
        # This will be a several-step operation.
        #
        # We first take all of our particles and assign them to Octs.  If they
        # are not in an Oct, we will assume they are out of bounds.  Note that
        # this means that if we have loaded neighbor particles for which an Oct
        # does not exist, we are going to be discarding them -- so sparse
        # octrees will need to ensure that neighbor octs *exist*.  Particles
        # will be assigned in a new NumPy array.  Note that this incurs
        # overhead, but reduces complexity as we will now be able to use
        # argsort.
        #
        # After the particles have been assigned to Octs, we process each Oct
        # individually.  We will do this by calling "get" for the *first*
        # particle in each set of Octs in the sorted list.  After this, we get
        # neighbors for each Oct.
        #
        # Now, with the set of neighbors (and thus their indices) we allocate
        # an array of particles and their fields, fill these in, and call our
        # process function.
        #
        # This is not terribly efficient -- for starters, the neighbor function
        # is not the most efficient yet.  We will also need to handle some
        # mechanism of an expandable array for holding pointers to Octs, so
        # that we can deal with >27 neighbors.
        if particle_octree is None:
            particle_octree = mesh_octree
            pdom_ind = mdom_ind
        cdef int nf, i, j, n
        cdef int dims[3]
        cdef np.float64_t[:] *field_check
        cdef np.float64_t **field_pointers
        cdef np.float64_t *field_vals
        cdef np.float64_t pos[3]
        cdef np.float64_t dds[3]
        cdef np.float64_t **octree_field_pointers
        cdef int nsize = 0
        cdef np.int64_t *nind = NULL
        cdef OctInfo moi, poi
        cdef Oct *oct
        cdef np.int64_t numpart, offset, local_ind, poff
        cdef np.int64_t moff_p, moff_m
        cdef np.int64_t[:] pind, doff, pdoms, pcount
        cdef np.int64_t[:,:] doff_m
        cdef np.ndarray[np.float64_t, ndim=1] tarr
        cdef np.ndarray[np.float64_t, ndim=4] iarr
        cdef np.float64_t[:,:] cart_positions
        cdef np.float64_t[:,:] oct_left_edges, oct_dds
        cdef OctInfo oinfo
        if geometry == "cartesian":
            self.pos_setup = cart_coord_setup
            cart_positions = positions
        elif geometry == "spherical":
            self.pos_setup = spherical_coord_setup
            cart_positions = np.empty((positions.shape[0], 3), dtype="float64")

            cart_positions[:,0] = positions[:,0] * \
                                  np.sin(positions[:,1]) * \
                                  np.cos(positions[:,2])
            cart_positions[:,1] = positions[:,0] * \
                                  np.sin(positions[:,1]) * \
                                  np.sin(positions[:,2])
            cart_positions[:,2] = positions[:,0] * \
                                  np.cos(positions[:,1])
            periodicity = (False, False, False)
        else:
            raise NotImplementedError
        dims[0] = dims[1] = dims[2] = mesh_octree.nz
        cdef int nz = dims[0] * dims[1] * dims[2]
        numpart = positions.shape[0]
        # pcount is the number of particles per oct.
        pcount = np.zeros_like(pdom_ind)
        oct_left_edges = np.zeros((pdom_ind.shape[0], 3), dtype='float64')
        oct_dds = np.zeros_like(oct_left_edges)
        # doff is the offset to a given oct in the sorted particles.
        doff = np.zeros_like(pdom_ind) - 1
        doff_m = np.zeros((mdom_ind.shape[0], 2), dtype="int64")
        moff_p = particle_octree.get_domain_offset(domain_id + domain_offset)
        moff_m = mesh_octree.get_domain_offset(domain_id + domain_offset)
        # pdoms points particles at their octs.  So the value in this array, for
        # a given index, is the local oct index.
        pdoms = np.zeros(positions.shape[0], dtype="int64") - 1
        nf = len(fields)
        if fields is None:
            fields = []
        field_pointers = <np.float64_t**> alloca(sizeof(np.float64_t *) * nf)
        for i in range(nf):
            tarr = fields[i]
            field_pointers[i] = <np.float64_t *> tarr.data
        if index_fields is None:
            index_fields = []
        nf = len(index_fields)
        index_field_pointers = <np.float64_t**> alloca(sizeof(np.float64_t *) * nf)
        for i in range(nf):
            iarr = index_fields[i]
            index_field_pointers[i] = <np.float64_t *> iarr.data
        for i in range(3):
            self.DW[i] = (mesh_octree.DRE[i] - mesh_octree.DLE[i])
            self.periodicity[i] = periodicity[i]
        cdef np.float64_t factor = particle_octree.nz
        for i in range(positions.shape[0]):
            for j in range(3):
                pos[j] = positions[i, j]
            oct = particle_octree.get(pos, &oinfo)
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            # Note that this has to be our local index, not our in-file index.
            # This is the particle count, which we'll use once we have sorted
            # the particles to calculate the offsets into each oct's particles.
            offset = oct.domain_ind - moff_p
            pcount[offset] += 1
            pdoms[i] = offset # We store the *actual* offset.
            # store oct positions and dds to avoid searching for neighbors
            # in octs that we know are too far away
            for j in range(3):
                oct_left_edges[offset, j] = oinfo.left_edge[j]
                oct_dds[offset, j] = oinfo.dds[j] * factor
        # Now we have oct assignments.  Let's sort them.
        # Note that what we will be providing to our processing functions will
        # actually be indirectly-sorted fields.  This preserves memory at the
        # expense of additional pointer lookups.
        pind = np.asarray(np.argsort(pdoms), dtype='int64', order='C')
        # So what this means is that we now have all the oct-0 particle indices
        # in order, then the oct-1, etc etc.
        # This now gives us the indices to the particles for each domain.
        for i in range(positions.shape[0]):
            # This value, poff, is the index of the particle in the *unsorted*
            # arrays.
            poff = pind[i]
            offset = pdoms[poff]
            # If we have yet to assign the starting index to this oct, we do so
            # now.
            if doff[offset] < 0: doff[offset] = i
        #print(domain_id, domain_offset, moff_p, moff_m)
        #raise RuntimeError
        # Now doff is full of offsets to the first entry in the pind that
        # refers to that oct's particles.
        cdef np.ndarray[np.uint8_t, ndim=1] visited
        visited = np.zeros(mdom_ind.shape[0], dtype="uint8")
        cdef int nproc = 0
        # This should be thread-private if we ever go to OpenMP
        cdef DistanceQueue dist_queue = DistanceQueue(self.maxn)
        dist_queue._setup(self.DW, self.periodicity)
        for i in range(oct_positions.shape[0]):
            if (i % 10000) == 0:
                PyErr_CheckSignals()
            for j in range(3):
                pos[j] = oct_positions[i, j]
            oct = mesh_octree.get(pos, &moi)
            offset = mdom_ind[oct.domain_ind - moff_m] * nz
            if visited[oct.domain_ind - moff_m] == 1: continue
            visited[oct.domain_ind - moff_m] = 1
            if offset < 0: continue
            nproc += 1
            self.neighbor_process(
                dims, moi.left_edge, moi.dds, cart_positions, field_pointers, doff,
                &nind, pind, pcount, offset, index_field_pointers,
                particle_octree, domain_id, &nsize, oct_left_edges,
                oct_dds, dist_queue)
        #print("VISITED", visited.sum(), visited.size,)
        #print(100.0*float(visited.sum())/visited.size)
        if nind != NULL:
            free(nind)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    def process_particles(self, OctreeContainer particle_octree,
                     np.ndarray[np.int64_t, ndim=1] pdom_ind,
                     np.ndarray[np.float64_t, ndim=2] positions,
                     fields = None, int domain_id = -1,
                     int domain_offset = 0,
                     periodicity = (True, True, True),
                     geometry = "cartesian"):
        # The other functions in this base class process particles in a way
        # that results in a modification to the *mesh*.  This function is
        # designed to process neighboring particles in such a way that a new
        # *particle* field is defined -- this means that new particle
        # attributes (*not* mesh attributes) can be created that rely on the
        # values of nearby particles.  For instance, a smoothing kernel, or a
        # nearest-neighbor field.
        cdef int nf, i, j, k, n
        cdef int dims[3]
        cdef np.float64_t **field_pointers
        cdef np.float64_t *field_vals
        cdef np.float64_t dds[3]
        cdef np.float64_t pos[3]
        cdef np.float64_t **octree_field_pointers
        cdef int nsize = 0
        cdef np.int64_t *nind = NULL
        cdef OctInfo moi, poi
        cdef Oct *oct
        cdef Oct **neighbors = NULL
        cdef np.int64_t nneighbors, numpart, offset, local_ind
        cdef np.int64_t moff_p, moff_m, pind0, poff
        cdef np.int64_t[:] pind, doff, pdoms, pcount
        cdef np.ndarray[np.float64_t, ndim=1] tarr
        cdef np.ndarray[np.float64_t, ndim=2] cart_positions
        if geometry == "cartesian":
            self.pos_setup = cart_coord_setup
            cart_positions = positions
        elif geometry == "spherical":
            self.pos_setup = spherical_coord_setup
            cart_positions = np.empty((positions.shape[0], 3), dtype="float64")

            cart_positions[:,0] = positions[:,0] * \
                                  np.sin(positions[:,1]) * \
                                  np.cos(positions[:,2])
            cart_positions[:,1] = positions[:,0] * \
                                  np.sin(positions[:,1]) * \
                                  np.sin(positions[:,2])
            cart_positions[:,2] = positions[:,0] * \
                                  np.cos(positions[:,1])
            periodicity = (False, False, False)
        else:
            raise NotImplementedError
        numpart = positions.shape[0]
        pcount = np.zeros_like(pdom_ind)
        doff = np.zeros_like(pdom_ind) - 1
        moff_p = particle_octree.get_domain_offset(domain_id + domain_offset)
        pdoms = np.zeros(positions.shape[0], dtype="int64") - 1
        nf = len(fields)
        if fields is None:
            fields = []
        field_pointers = <np.float64_t**> alloca(sizeof(np.float64_t *) * nf)
        for i in range(nf):
            tarr = fields[i]
            field_pointers[i] = <np.float64_t *> tarr.data
        for i in range(3):
            self.DW[i] = (particle_octree.DRE[i] - particle_octree.DLE[i])
            self.periodicity[i] = periodicity[i]
        for i in range(positions.shape[0]):
            for j in range(3):
                pos[j] = positions[i, j]
            oct = particle_octree.get(pos)
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            # Note that this has to be our local index, not our in-file index.
            # This is the particle count, which we'll use once we have sorted
            # the particles to calculate the offsets into each oct's particles.
            offset = oct.domain_ind - moff_p
            pcount[offset] += 1
            pdoms[i] = offset # We store the *actual* offset.
        # Now we have oct assignments.  Let's sort them.
        # Note that what we will be providing to our processing functions will
        # actually be indirectly-sorted fields.  This preserves memory at the
        # expense of additional pointer lookups.
        pind = np.asarray(np.argsort(pdoms), dtype='int64', order='C')
        # So what this means is that we now have all the oct-0 particle indices
        # in order, then the oct-1, etc etc.
        # This now gives us the indices to the particles for each domain.
        for i in range(positions.shape[0]):
            # This value, poff, is the index of the particle in the *unsorted*
            # arrays.
            poff = pind[i]
            offset = pdoms[poff]
            # If we have yet to assign the starting index to this oct, we do so
            # now.
            if doff[offset] < 0: doff[offset] = i
        #print(domain_id, domain_offset, moff_p, moff_m)
        #raise RuntimeError
        # Now doff is full of offsets to the first entry in the pind that
        # refers to that oct's particles.
        cdef int maxnei = 0
        cdef int nproc = 0
        # This should be thread-private if we ever go to OpenMP
        cdef DistanceQueue dist_queue = DistanceQueue(self.maxn)
        dist_queue._setup(self.DW, self.periodicity)
        for i in range(doff.shape[0]):
            if doff[i] < 0: continue
            offset = pind[doff[i]]
            for j in range(3):
                pos[j] = positions[offset, j]
            for j in range(pcount[i]):
                pind0 = pind[doff[i] + j]
                for k in range(3):
                    pos[k] = positions[pind0, k]
                self.neighbor_process_particle(pos, cart_positions, field_pointers,
                            doff, &nind, pind, pcount, pind0,
                            NULL, particle_octree, domain_id, &nsize,
                            dist_queue)
        #print("VISITED", visited.sum(), visited.size,)
        #print(100.0*float(visited.sum())/visited.size)
        if nind != NULL:
            free(nind)

    cdef int neighbor_search(self, np.float64_t pos[3], OctreeContainer octree,
                             np.int64_t **nind, int *nsize,
                             np.int64_t nneighbors, np.int64_t domain_id,
                             Oct **oct = NULL, int extra_layer = 0):
        cdef OctInfo oi
        cdef Oct *ooct
        cdef Oct **neighbors
        cdef Oct **first_layer
        cdef int j, total_neighbors = 0, initial_layer = 0
        cdef int layer_ind = 0
        cdef np.int64_t moff = octree.get_domain_offset(domain_id)
        ooct = octree.get(pos, &oi)
        if oct != NULL and ooct == oct[0]:
            return nneighbors
        oct[0] = ooct
        if nind[0] == NULL:
            nsize[0] = 27
            nind[0] = <np.int64_t *> malloc(sizeof(np.int64_t)*nsize[0])
        # This is our "seed" set of neighbors.  If we are asked to, we will
        # create a master list of neighbors that is much bigger and includes
        # everything.
        layer_ind = 0
        first_layer = NULL
        while 1:
            neighbors = octree.neighbors(&oi, &nneighbors, ooct, self.periodicity)
            # Now we have all our neighbors.  And, we should be set for what
            # else we need to do.
            if total_neighbors + nneighbors > nsize[0]:
                nind[0] = <np.int64_t *> realloc(
                    nind[0], sizeof(np.int64_t)*(nneighbors + total_neighbors))
                nsize[0] = nneighbors + total_neighbors
            for j in range(nneighbors):
                # Particle octree neighbor indices
                nind[0][j + total_neighbors] = neighbors[j].domain_ind - moff
            total_neighbors += nneighbors
            if extra_layer == 0:
                # Not adding on any additional layers here.
                free(neighbors)
                neighbors = NULL
                break
            if initial_layer == 0:
                initial_layer = nneighbors
                first_layer = neighbors
            else:
                # Allocated internally; we free this in the loops if we aren't
                # tracking it
                free(neighbors)
                neighbors = NULL
            ooct = first_layer[layer_ind]
            layer_ind += 1
            if layer_ind == initial_layer:
                neighbors
                break


        for j in range(total_neighbors):
            # Particle octree neighbor indices
            if nind[0][j] == -1: continue
            for n in range(j):
                if nind[0][j] == nind[0][n]:
                    nind[0][j] = -1
        # This is allocated by the neighbors function, so we deallocate it.
        if first_layer != NULL:
            free(first_layer)
        return total_neighbors

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    def process_grid(self, gobj,
                     np.ndarray[np.float64_t, ndim=2] positions,
                     fields = None):
        raise NotImplementedError

    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **ifields, DistanceQueue dq):
        raise NotImplementedError

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void neighbor_find(self,
                            np.int64_t nneighbors,
                            np.int64_t *nind,
                            np.int64_t[:] doffs,
                            np.int64_t[:] pcounts,
                            np.int64_t[:] pinds,
                            np.float64_t[:,:] ppos,
                            np.float64_t cpos[3],
                            np.float64_t[:,:] oct_left_edges,
                            np.float64_t[:,:] oct_dds,
                            DistanceQueue dq
                            ):
        # We are now given the number of neighbors, the indices into the
        # domains for them, and the number of particles for each.
        cdef int ni, i, j, k
        cdef np.int64_t offset, pn, pc
        cdef np.float64_t pos[3]
        cdef np.float64_t ex[2]
        cdef np.float64_t DR[2]
        cdef np.float64_t cp, r2_trunc, r2, dist
        dq.neighbor_reset()
        for ni in range(nneighbors):
            if nind[ni] == -1: continue
            # terminate early if all 8 corners of oct are farther away than
            # most distant currently known neighbor
            if oct_left_edges != None and dq.curn == dq.maxn:
                r2_trunc = dq.neighbors[dq.curn - 1].r2
                # iterate over each dimension in the outer loop so we can
                # consolidate temporary storage
                # What this next bit does is figure out which component is the
                # closest, of each possible permutation.
                # k here is the dimension
                r2 = 0.0
                for k in range(3):
                    # We start at left edge, then do halfway, then right edge.
                    ex[0] = oct_left_edges[nind[ni], k]
                    ex[1] = ex[0] + oct_dds[nind[ni], k]
                    # There are three possibilities; we are between, left-of,
                    # or right-of the extrema.  Thanks to
                    # http://stackoverflow.com/questions/5254838/calculating-distance-between-a-point-and-a-rectangular-box-nearest-point
                    # for some help.  This has been modified to account for
                    # periodicity.
                    dist = 0.0
                    DR[0] = (ex[0] - cpos[k])
                    DR[1] = (cpos[k] - ex[1])
                    for j in range(2):
                        if not self.periodicity[k]:
                            pass
                        elif (DR[j] > self.DW[k]/2.0):
                            DR[j] -= self.DW[k]
                        elif (DR[j] < -self.DW[k]/2.0):
                            DR[j] += self.DW[k]
                        dist = fmax(dist, DR[j])
                    r2 += dist*dist
                if r2 > r2_trunc:
                    continue
            offset = doffs[nind[ni]]
            pc = pcounts[nind[ni]]
            for i in range(pc):
                pn = pinds[offset + i]
                for j in range(3):
                    pos[j] = ppos[pn, j]
                dq.neighbor_eval(pn, pos, cpos)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void neighbor_process(self, int dim[3], np.float64_t left_edge[3],
                               np.float64_t dds[3], np.float64_t[:,:] ppos,
                               np.float64_t **fields,
                               np.int64_t [:] doffs, np.int64_t **nind,
                               np.int64_t [:] pinds, np.int64_t[:] pcounts,
                               np.int64_t offset,
                               np.float64_t **index_fields,
                               OctreeContainer octree, np.int64_t domain_id,
                               int *nsize, np.float64_t[:,:] oct_left_edges,
                               np.float64_t[:,:] oct_dds,
                               DistanceQueue dq):
        # Note that we assume that fields[0] == smoothing length in the native
        # units supplied.  We can now iterate over every cell in the block and
        # every particle to find the nearest.  We will use a priority heap.
        cdef int i, j, k, ntot, nntot, m, nneighbors
        cdef np.float64_t cpos[3]
        cdef np.float64_t opos[3]
        cdef Oct* oct = NULL
        cpos[0] = left_edge[0] + 0.5*dds[0]
        for i in range(dim[0]):
            cpos[1] = left_edge[1] + 0.5*dds[1]
            for j in range(dim[1]):
                cpos[2] = left_edge[2] + 0.5*dds[2]
                for k in range(dim[2]):
                    self.pos_setup(cpos, opos)
                    nneighbors = self.neighbor_search(opos, octree,
                                    nind, nsize, nneighbors, domain_id, &oct, 0)
                    self.neighbor_find(nneighbors, nind[0], doffs, pcounts,
                                       pinds, ppos, opos, oct_left_edges,
                                       oct_dds, dq)
                    # Now we have all our neighbors in our neighbor list.
                    if dq.curn <-1*dq.maxn:
                        ntot = nntot = 0
                        for m in range(nneighbors):
                            if nind[0][m] < 0: continue
                            nntot += 1
                            ntot += pcounts[nind[0][m]]
                        print("SOMETHING WRONG", dq.curn, nneighbors, ntot, nntot)
                    self.process(offset, i, j, k, dim, opos, fields,
                                 index_fields, dq)
                    cpos[2] += dds[2]
                cpos[1] += dds[1]
            cpos[0] += dds[0]

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void neighbor_process_particle(self, np.float64_t cpos[3],
                               np.float64_t[:,:] ppos,
                               np.float64_t **fields,
                               np.int64_t[:] doffs, np.int64_t **nind,
                               np.int64_t[:] pinds, np.int64_t[:] pcounts,
                               np.int64_t offset,
                               np.float64_t **index_fields,
                               OctreeContainer octree,
                               np.int64_t domain_id, int *nsize,
                               DistanceQueue dq):
        # Note that we assume that fields[0] == smoothing length in the native
        # units supplied.  We can now iterate over every cell in the block and
        # every particle to find the nearest.  We will use a priority heap.
        cdef int i, j, k, ntot, nntot, m
        cdef int dim[3]
        cdef Oct *oct = NULL
        cdef np.int64_t nneighbors = 0
        i = j = k = 0
        dim[0] = dim[1] = dim[2] = 1
        cdef np.float64_t opos[3]
        self.pos_setup(cpos, opos)
        nneighbors = self.neighbor_search(opos, octree,
                        nind, nsize, nneighbors, domain_id, &oct, 0)
        self.neighbor_find(nneighbors, nind[0], doffs, pcounts, pinds, ppos,
                           opos, None, None, dq)
        self.process(offset, i, j, k, dim, opos, fields, index_fields, dq)

cdef class VolumeWeightedSmooth(ParticleSmoothOperation):
    # This smoothing function evaluates the field, *without* normalization, at
    # every point in the *mesh*.  Applying a normalization results in
    # non-conservation of mass when smoothing density; to avoid this, we do not
    # apply this normalization factor.  The SPLASH paper
    # (http://arxiv.org/abs/0709.0832v1) discusses this in detail; what we are
    # applying here is equation 6, with variable smoothing lengths (eq 13).
    cdef np.float64_t **fp
    cdef public object vals
    def initialize(self):
        cdef int i
        if self.nfields < 4:
            # We need four fields -- the mass should be the first, then the
            # smoothing length for particles, the normalization factor to
            # ensure mass conservation, then the field we're smoothing.
            raise RuntimeError
        cdef np.ndarray tarr
        self.fp = <np.float64_t **> malloc(
            sizeof(np.float64_t *) * (self.nfields - 3))
        self.vals = []
        # We usually only allocate one field; if we are doing multiple field,
        # single-pass smoothing, then we might have more.
        for i in range(self.nfields - 3):
            tarr = np.zeros(self.nvals, dtype="float64", order="F")
            self.vals.append(tarr)
            self.fp[i] = <np.float64_t *> tarr.data

    def finalize(self):
        free(self.fp)
        return self.vals

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq):
        # We have our i, j, k for our cell, as well as the cell position.
        # We also have a list of neighboring particles with particle numbers.
        cdef int n, fi
        cdef np.float64_t weight, r2, val, hsml, dens, mass, coeff, max_r
        cdef np.float64_t max_hsml, ihsml, ihsml3, kern
        coeff = 0.0
        cdef np.int64_t pn
        # We get back our mass
        # rho_i = sum(j = 1 .. n) m_j * W_ij
        max_r = sqrt(dq.neighbors[dq.curn-1].r2)
        max_hsml = index_fields[0][gind(i,j,k,dim) + offset]
        for n in range(dq.curn):
            # No normalization for the moment.
            # fields[0] is the smoothing length.
            r2 = dq.neighbors[n].r2
            pn = dq.neighbors[n].pn
            # Smoothing kernel weight function
            mass = fields[0][pn]
            hsml = fields[1][pn]
            dens = fields[2][pn]
            if hsml < 0:
                hsml = max_r
            if hsml == 0: continue
            ihsml = 1.0/hsml
            hsml = fmax(max_hsml/2.0, hsml)
            ihsml3 = ihsml*ihsml*ihsml
            # Usually this density has been computed
            if dens == 0.0: continue
            weight = (mass / dens) * ihsml3
            kern = self.sph_kernel(sqrt(r2) * ihsml)
            weight *= kern
            # Mass of the particle times the value
            for fi in range(self.nfields - 3):
                val = fields[fi + 3][pn]
                self.fp[fi][gind(i,j,k,dim) + offset] += val * weight
        return

volume_weighted_smooth = VolumeWeightedSmooth

cdef class NearestNeighborSmooth(ParticleSmoothOperation):
    cdef np.float64_t *fp
    cdef public object vals
    def initialize(self):
        cdef np.ndarray tarr
        assert(self.nfields == 1)
        tarr = np.zeros(self.nvals, dtype="float64", order="F")
        self.vals = tarr
        self.fp = <np.float64_t *> tarr.data

    def finalize(self):
        return self.vals

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq):
        # We have our i, j, k for our cell, as well as the cell position.
        # We also have a list of neighboring particles with particle numbers.
        cdef np.int64_t pn
        # We get back our mass
        # rho_i = sum(j = 1 .. n) m_j * W_ij
        pn = dq.neighbors[0].pn
        self.fp[gind(i,j,k,dim) + offset] = fields[0][pn]
        #self.fp[gind(i,j,k,dim) + offset] = dq.neighbors[0].r2
        return

nearest_smooth = NearestNeighborSmooth

cdef class IDWInterpolationSmooth(ParticleSmoothOperation):
    cdef np.float64_t *fp
    cdef public int p2
    cdef public object vals
    def initialize(self):
        cdef np.ndarray tarr
        assert(self.nfields == 1)
        tarr = np.zeros(self.nvals, dtype="float64", order="F")
        self.vals = tarr
        self.fp = <np.float64_t *> tarr.data
        self.p2 = 2 # Power, for IDW, in units of 2.  So we only do even p's.

    def finalize(self):
        return self.vals

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq):
        # We have our i, j, k for our cell, as well as the cell position.
        # We also have a list of neighboring particles with particle numbers.
        cdef np.int64_t pn, ni, di
        cdef np.float64_t total_weight = 0.0, total_value = 0.0, r2, val, w
        # We're going to do a very simple IDW average
        if dq.neighbors[0].r2 == 0.0:
            pn = dq.neighbors[0].pn
            self.fp[gind(i,j,k,dim) + offset] = fields[0][pn]
        for ni in range(dq.curn):
            r2 = dq.neighbors[ni].r2
            val = fields[0][dq.neighbors[ni].pn]
            w = r2
            for di in range(self.p2 - 1):
                w *= r2
            total_value += w * val
            total_weight += w
        self.fp[gind(i,j,k,dim) + offset] = total_value / total_weight
        return

idw_smooth = IDWInterpolationSmooth

cdef class NthNeighborDistanceSmooth(ParticleSmoothOperation):

    def initialize(self):
        return

    def finalize(self):
        return

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq):
        cdef np.float64_t max_r
        # We assume "offset" here is the particle index.
        max_r = sqrt(dq.neighbors[dq.curn-1].r2)
        fields[0][offset] = max_r

nth_neighbor_smooth = NthNeighborDistanceSmooth

cdef class SmoothedDensityEstimate(ParticleSmoothOperation):
    def initialize(self):
        return

    def finalize(self):
        return

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void process(self, np.int64_t offset, int i, int j, int k,
                      int dim[3], np.float64_t cpos[3], np.float64_t **fields,
                      np.float64_t **index_fields, DistanceQueue dq):
        cdef np.float64_t r2, hsml, dens, mass, weight, lw
        cdef int pn
        # We assume "offset" here is the particle index.
        hsml = sqrt(dq.neighbors[dq.curn-1].r2)
        dens = 0.0
        weight = 0.0
        for pn in range(dq.curn):
            mass = fields[0][dq.neighbors[pn].pn]
            r2 = dq.neighbors[pn].r2
            lw = self.sph_kernel(sqrt(r2) / hsml)
            dens += mass * lw
        weight = (4.0/3.0) * 3.1415926 * hsml**3
        fields[1][offset] = dens/weight

density_smooth = SmoothedDensityEstimate



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Particle Deposition onto Cells




"""


cimport numpy as np

import numpy as np

cimport cython
from cython.view cimport memoryview as cymemview
from libc.math cimport sqrt
from libc.stdlib cimport free, malloc
from oct_container cimport Oct, OctInfo, OctreeContainer

from yt.utilities.lib.fp_utils cimport *

from yt.utilities.lib.misc_utilities import OnceIndirect


cdef append_axes(np.ndarray arr, int naxes):
    if arr.ndim == naxes:
        return arr
    # Avoid copies
    arr2 = arr.view()
    arr2.shape = arr2.shape + (1,) * (naxes - arr2.ndim)
    return arr2

cdef class ParticleDepositOperation:
    def __init__(self, nvals, kernel_name):
        # nvals is a tuple containing the active dimensions of the
        # grid to deposit onto and the number of grids,
        # (nx, ny, nz, ngrids)
        self.nvals = nvals
        self.update_values = 0 # This is the default
        self.sph_kernel = get_kernel_func(kernel_name)

    def initialize(self, *args):
        raise NotImplementedError

    def finalize(self, *args):
        raise NotImplementedError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def process_octree(self, OctreeContainer octree,
                     np.ndarray[np.int64_t, ndim=1] dom_ind,
                     np.ndarray[np.float64_t, ndim=2] positions,
                     fields = None, int domain_id = -1,
                     int domain_offset = 0, lvlmax = None):
        cdef int nf, i, j
        if fields is None:
            fields = []
        nf = len(fields)
        cdef np.float64_t[::cython.view.indirect, ::1] field_pointers
        if nf > 0: field_pointers = OnceIndirect(fields)
        cdef np.float64_t pos[3]
        cdef np.float64_t[:] field_vals = np.empty(nf, dtype="float64")
        cdef int dims[3]
        dims[0] = dims[1] = dims[2] = octree.nz
        cdef int nz = dims[0] * dims[1] * dims[2]
        cdef OctInfo oi
        cdef np.int64_t offset, moff
        cdef Oct *oct
        cdef np.int64_t numpart = positions.shape[0]
        cdef np.int8_t use_lvlmax
        moff = octree.get_domain_offset(domain_id + domain_offset)
        if lvlmax is None:
            use_lvlmax = False
            lvlmax = []
        else:
            use_lvlmax = True
        cdef np.ndarray[np.int32_t, ndim=1] lvlmaxval = np.asarray(lvlmax, dtype=np.int32)

        for i in range(positions.shape[0]):
            # We should check if particle remains inside the Oct here
            for j in range(nf):
                field_vals[j] = field_pointers[j,i]
            for j in range(3):
                pos[j] = positions[i, j]
            # This line should be modified to have it return the index into an
            # array based on whatever cutting of the domain we have done.  This
            # may or may not include the domain indices that we have
            # previously generated.  This way we can support not knowing the
            # full octree structure.  All we *really* care about is some
            # arbitrary offset into a field value for deposition.
            if not use_lvlmax:
                oct = octree.get(pos, &oi)
            else:
                oct = octree.get(pos, &oi, max_level=lvlmaxval[i])
            # This next line is unfortunate.  Basically it says, sometimes we
            # might have particles that belong to octs outside our domain.
            # For the distributed-memory octrees, this will manifest as a NULL
            # oct.  For the non-distributed memory octrees, we'll simply see
            # this as a domain_id that is not the current domain id.  Note that
            # this relies on the idea that all the particles in a region are
            # all fed to sequential domain subsets, which will not be true with
            # RAMSES, where we *will* miss particles that live in ghost
            # regions on other processors.  Addressing this is on the TODO
            # list.
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            # Note that this has to be our local index, not our in-file index.
            offset = dom_ind[oct.domain_ind - moff]
            if offset < 0: continue
            # Check that we found the oct ...
            self.process(dims, i, oi.left_edge, oi.dds,
                         offset, pos, field_vals, oct.domain_ind)
            if self.update_values == 1:
                for j in range(nf):
                    field_pointers[j][i] = field_vals[j]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def process_grid(self, gobj,
                     np.ndarray[np.float64_t, ndim=2, cast=True] positions,
                     fields = None):
        cdef int nf, i, j
        if fields is None:
            fields = []
        if positions.shape[0] == 0: return
        nf = len(fields)
        cdef np.float64_t[:] field_vals = np.empty(nf, dtype="float64")
        cdef np.float64_t[::cython.view.indirect, ::1] field_pointers
        if nf > 0: field_pointers = OnceIndirect(fields)
        cdef np.float64_t pos[3]
        cdef np.int64_t gid = getattr(gobj, "id", -1)
        cdef np.float64_t dds[3]
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef int dims[3]
        for i in range(3):
            dds[i] = gobj.dds[i]
            left_edge[i] = gobj.LeftEdge[i]
            right_edge[i] = gobj.RightEdge[i]
            dims[i] = gobj.ActiveDimensions[i]
        for i in range(positions.shape[0]):
            # Now we process
            for j in range(nf):
                field_vals[j] = field_pointers[j,i]
            for j in range(3):
                pos[j] = positions[i, j]
            continue_loop = False
            for j in range(3):
                if pos[j] < left_edge[j] or pos[j] > right_edge[j]:
                    continue_loop = True
            if continue_loop:
                continue
            self.process(dims, i, left_edge, dds, 0, pos, field_vals, gid)
            if self.update_values == 1:
                for j in range(nf):
                    field_pointers[j][i] = field_vals[j]

    cdef int process(self, int dim[3], int ipart, np.float64_t left_edge[3],
                     np.float64_t dds[3], np.int64_t offset,
                     np.float64_t ppos[3], np.float64_t[:] fields,
                     np.int64_t domain_ind) nogil except -1:
        with gil:
            raise NotImplementedError

cdef class CountParticles(ParticleDepositOperation):
    cdef np.int64_t[:,:,:,:] count
    def initialize(self):
        # Create a numpy array accessible to python
        self.count = append_axes(
            np.zeros(self.nvals, dtype="int64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset, # offset into IO field
                     np.float64_t ppos[3], # this particle's position
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        # here we do our thing; this is the kernel
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
        self.count[ii[2], ii[1], ii[0], offset] += 1
        return 0

    def finalize(self):
        arr = np.asarray(self.count)
        arr.shape = self.nvals
        return arr.astype("float64")

deposit_count = CountParticles

cdef class SimpleSmooth(ParticleDepositOperation):
    # Note that this does nothing at the edges.  So it will give a poor
    # estimate there, and since Octrees are mostly edges, this will be a very
    # poor SPH kernel.
    cdef np.float64_t[:,:,:,:] data
    cdef np.float64_t[:,:,:,:] temp

    def initialize(self):
        self.data = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.temp = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        cdef int ii[3]
        cdef int ib0[3]
        cdef int ib1[3]
        cdef int i, j, k, half_len
        cdef np.float64_t idist[3]
        cdef np.float64_t kernel_sum, dist
        # Smoothing length is fields[0]
        kernel_sum = 0.0
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
            half_len = <int>(fields[0]/dds[i]) + 1
            ib0[i] = ii[i] - half_len
            ib1[i] = ii[i] + half_len
            if ib0[i] >= dim[i] or ib1[i] <0:
                return 0
            ib0[i] = iclip(ib0[i], 0, dim[i] - 1)
            ib1[i] = iclip(ib1[i], 0, dim[i] - 1)
        for i from ib0[0] <= i <= ib1[0]:
            idist[0] = (ii[0] - i) * dds[0]
            idist[0] *= idist[0]
            for j from ib0[1] <= j <= ib1[1]:
                idist[1] = (ii[1] - j) * dds[1]
                idist[1] *= idist[1]
                for k from ib0[2] <= k <= ib1[2]:
                    idist[2] = (ii[2] - k) * dds[2]
                    idist[2] *= idist[2]
                    dist = idist[0] + idist[1] + idist[2]
                    # Calculate distance in multiples of the smoothing length
                    dist = sqrt(dist) / fields[0]
                    with gil:
                        self.temp[k,j,i,offset] = self.sph_kernel(dist)
                    kernel_sum += self.temp[k,j,i,offset]
        # Having found the kernel, deposit accordingly into gdata
        for i from ib0[0] <= i <= ib1[0]:
            for j from ib0[1] <= j <= ib1[1]:
                for k from ib0[2] <= k <= ib1[2]:
                    dist = self.temp[k,j,i,offset] / kernel_sum
                    self.data[k,j,i,offset] += fields[1] * dist
        return 0

    def finalize(self):
        return self.odata

deposit_simple_smooth = SimpleSmooth

cdef class SumParticleField(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] sum
    def initialize(self):
        self.sum = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i]) / dds[i])
        self.sum[ii[2], ii[1], ii[0], offset] += fields[0]
        return 0

    def finalize(self):
        sum = np.asarray(self.sum)
        sum.shape = self.nvals
        return sum

deposit_sum = SumParticleField

cdef class StdParticleField(ParticleDepositOperation):
    # Thanks to Britton and MJ Turk for the link
    # to a single-pass STD
    # http://www.cs.berkeley.edu/~mhoemmen/cs194/Tutorials/variance.pdf
    cdef np.float64_t[:,:,:,:] mk
    cdef np.float64_t[:,:,:,:] qk
    cdef np.float64_t[:,:,:,:] i
    def initialize(self):
        # we do this in a single pass, but need two scalar
        # per cell, M_k, and Q_k and also the number of particles
        # deposited into each one
        # the M_k term
        self.mk = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.qk = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.i = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        cdef int ii[3]
        cdef int i, cell_index
        cdef float k, mk, qk
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
        k = self.i[ii[2], ii[1], ii[0], offset]
        mk = self.mk[ii[2], ii[1], ii[0], offset]
        qk = self.qk[ii[2], ii[1], ii[0], offset]
        if k == 0.0:
            # Initialize cell values
            self.mk[ii[2], ii[1], ii[0], offset] = fields[0]
        else:
            self.mk[ii[2], ii[1], ii[0], offset] = mk + (fields[0] - mk) / k
            self.qk[ii[2], ii[1], ii[0], offset] = \
                qk + (k - 1.0) * (fields[0] - mk) * (fields[0] - mk) / k
        self.i[ii[2], ii[1], ii[0], offset] += 1
        return 0

    def finalize(self):
        # This is the standard variance
        # if we want sample variance divide by (self.oi - 1.0)
        i = np.asarray(self.i)
        std2 = np.asarray(self.qk) / i
        std2[i == 0.0] = 0.0
        std2.shape = self.nvals
        return np.sqrt(std2)

deposit_std = StdParticleField

cdef class CICDeposit(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] field
    cdef public object ofield
    def initialize(self):
        if not all(_ > 1 for _ in self.nvals[:-1]):
            from yt.utilities.exceptions import YTBoundsDefinitionError
            raise YTBoundsDefinitionError(
                "CIC requires minimum of 2 zones in all spatial dimensions.",
                self.nvals[:-1])
        self.field = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset, # offset into IO field
                     np.float64_t ppos[3], # this particle's position
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:

        cdef int i, j, k
        cdef np.uint64_t ii
        cdef int ind[3]
        cdef np.float64_t rpos[3]
        cdef np.float64_t rdds[3][2]
        cdef np.float64_t fact, edge0, edge1, edge2
        cdef np.float64_t le0, le1, le2
        cdef np.float64_t dx, dy, dz, dx2, dy2, dz2

        # Compute the position of the central cell
        for i in range(3):
            rpos[i] = (ppos[i]-left_edge[i])/dds[i]
            rpos[i] = fclip(rpos[i], 0.5001, dim[i]-0.5001)
            ind[i] = <int> (rpos[i] + 0.5)
            # Note these are 1, then 0
            rdds[i][1] = (<np.float64_t> ind[i]) + 0.5 - rpos[i]
            rdds[i][0] = 1.0 - rdds[i][1]

        for i in range(2):
            for j in range(2):
                for k in range(2):
                    self.field[ind[2] - k, ind[1] - j, ind[0] - i, offset] += \
                        fields[0]*rdds[0][i]*rdds[1][j]*rdds[2][k]

        return 0

    def finalize(self):
        rv = np.asarray(self.field)
        rv.shape = self.nvals
        return rv

deposit_cic = CICDeposit

cdef class WeightedMeanParticleField(ParticleDepositOperation):
    # Deposit both mass * field and mass into two scalars
    # then in finalize divide mass * field / mass
    cdef np.float64_t[:,:,:,:] wf
    cdef np.float64_t[:,:,:,:] w
    def initialize(self):
        self.wf = append_axes(
            np.zeros(self.nvals, dtype='float64', order='F'), 4)
        self.w = append_axes(
            np.zeros(self.nvals, dtype='float64', order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i]) / dds[i])
        self.w[ii[2], ii[1], ii[0], offset] += fields[1]
        self.wf[ii[2], ii[1], ii[0], offset] += fields[0] * fields[1]
        return 0

    def finalize(self):
        wf = np.asarray(self.wf)
        w = np.asarray(self.w)
        with np.errstate(divide='ignore', invalid='ignore'):
            rv = wf / w
        rv.shape = self.nvals
        return rv

deposit_weighted_mean = WeightedMeanParticleField

cdef class MeshIdentifier(ParticleDepositOperation):
    # This is a tricky one!  What it does is put into the particle array the
    # value of the oct or block (grids will always be zero) identifier that a
    # given particle resides in
    def initialize(self):
        self.update_values = 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                      np.float64_t left_edge[3],
                      np.float64_t dds[3],
                      np.int64_t offset,
                      np.float64_t ppos[3],
                      np.float64_t[:] fields,
                      np.int64_t domain_ind
                      ) nogil except -1:
        fields[0] = domain_ind
        return 0

    def finalize(self):
        return

deposit_mesh_id = MeshIdentifier

cdef class CellIdentifier(ParticleDepositOperation):
    cdef np.int64_t[:] indexes, cell_index
    # This method stores the offset of the grid containing each particle
    # and compute the index of the cell in the oct.
    def initialize(self, int npart):
        self.indexes = np.zeros(npart, dtype=np.int64, order='F') - 1
        self.cell_index = np.zeros(npart, dtype=np.int64, order='F') - 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                      np.float64_t left_edge[3],
                      np.float64_t dds[3],
                      np.int64_t offset,
                      np.float64_t ppos[3],
                      np.float64_t[:] fields,
                      np.int64_t domain_ind
                      ) nogil except -1:
        cdef int i, icell
        self.indexes[ipart] = offset

        icell = 0
        for i in range(3):
            if ppos[i] > left_edge[i] + dds[i]:
                icell |= 4 >> i

        # Compute cell index
        self.cell_index[ipart] = icell

        return 0

    def finalize(self):
        return np.asarray(self.indexes), np.asarray(self.cell_index)

deposit_cell_id = CellIdentifier

cdef class NNParticleField(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] nnfield
    cdef np.float64_t[:,:,:,:] distfield
    def initialize(self):
        self.nnfield = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.distfield = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.distfield[:] = np.inf

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) nogil except -1:
        # This one is a bit slow.  Every grid cell is going to be iterated
        # over, and we're going to deposit particles in it.
        cdef int i, j, k
        cdef int ii[3]
        cdef np.float64_t r2
        cdef np.float64_t gpos[3]
        gpos[0] = left_edge[0] + 0.5 * dds[0]
        for i in range(dim[0]):
            gpos[1] = left_edge[1] + 0.5 * dds[1]
            for j in range(dim[1]):
                gpos[2] = left_edge[2] + 0.5 * dds[2]
                for k in range(dim[2]):
                    r2 = ((ppos[0] - gpos[0])*(ppos[0] - gpos[0]) +
                          (ppos[1] - gpos[1])*(ppos[1] - gpos[1]) +
                          (ppos[2] - gpos[2])*(ppos[2] - gpos[2]))
                    if r2 < self.distfield[k,j,i,offset]:
                        self.distfield[k,j,i,offset] = r2
                        self.nnfield[k,j,i,offset] = fields[0]
                    gpos[2] += dds[2]
                gpos[1] += dds[1]
            gpos[0] += dds[0]
        return 0

    def finalize(self):
        nn = np.asarray(self.nnfield)
        nn.shape = self.nvals
        return nn

deposit_nearest = NNParticleField



import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Set the format of the ionozation_label to be `plus_minus`
# instead of the default `roman_numeral`
ds.set_field_label_format("ionization_label", "plus_minus")

slc = yt.SlicePlot(ds, "x", ("gas", "H_p1_number_density"))
slc.save("plus_minus_ionization_format_sliceplot.png")
